# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Authentication tests for Nessie CLI."""

from pynessie.error import (
    NessieConflictException,
    NessieContentNotFoundException,
    NessieException,
    NessieNotFoundException,
    NessiePermissionException,
    NessiePreconidtionFailedException,
    NessieReferenceAlreadyExistsException,
    NessieReferenceConflictException,
    NessieReferenceNotFoundException,
    NessieServerException,
    NessieUnauthorizedException,
    _create_exception,
)

# Non-VCR tests for error response handling


def test_raise_exception_missing_payload() -> None:
    """Test the handling error responses with missing JSON payload."""
    ex = _create_exception({}, 412, "reason123", "url123")
    assert isinstance(ex, NessiePreconidtionFailedException)
    assert "412" in str(ex.json())
    assert "reason123" in str(ex.json())
    assert "url123" in str(ex.json())

    ex = _create_exception({}, 401, "reason123", "url123")
    assert isinstance(ex, NessieUnauthorizedException)

    ex = _create_exception({}, 403, "reason123", "url123")
    assert isinstance(ex, NessiePermissionException)

    ex = _create_exception({}, 404, "reason123", "url123")
    assert isinstance(ex, NessieNotFoundException)

    ex = _create_exception({}, 409, "reason123", "url123")
    assert isinstance(ex, NessieConflictException)

    ex = _create_exception({}, 599, "reason123", "url123")
    assert isinstance(ex, NessieServerException)
    assert "599" in str(ex.json())
    assert "Server Error" in str(ex)
    assert "Server Error" in str(ex.json())
    assert "Internal Server Error" in str(ex.json())

    ex = _create_exception({}, 12345, "reason123", "url123")
    assert isinstance(ex, NessieException)
    assert "12345" in str(ex.json())


def _test_error_code(error_code: str, exception: type) -> None:
    error_dict = {
        "message": "msg123",
        "status": 499,
        "errorCode": error_code,
    }
    ex = _create_exception(error_dict, 498, "reason123", "url123")
    assert isinstance(ex, exception)
    ex_str = str(ex)
    assert "Client Error" in ex_str
    assert "msg123" in ex_str
    assert "reason123" in ex_str
    ex_str_json = str(ex.json())
    assert "Client Error" in ex_str_json
    assert "499" in ex_str_json
    assert "498" in ex_str_json
    assert "msg123" in ex_str_json
    assert "reason123" in ex_str_json
    assert "url123" in ex_str_json


def test_raise_ref_not_found() -> None:
    """Test the handling error code REFERENCE_NOT_FOUND."""
    _test_error_code("REFERENCE_NOT_FOUND", NessieReferenceNotFoundException)


def test_raise_content_not_found() -> None:
    """Test the handling error code CONTENT_NOT_FOUND."""
    _test_error_code("CONTENT_NOT_FOUND", NessieContentNotFoundException)


def test_raise_ref_conflict() -> None:
    """Test the handling error code REFERENCE_CONFLICT."""
    _test_error_code("REFERENCE_CONFLICT", NessieReferenceConflictException)


def test_raise_ref_already_exists() -> None:
    """Test the handling error code REFERENCE_ALREADY_EXISTS."""
    _test_error_code("REFERENCE_ALREADY_EXISTS", NessieReferenceAlreadyExistsException)



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""diff CLI command."""

import click

from pynessie.cli_common_context import ContextObject
from pynessie.decorators import error_handler, pass_client
from pynessie.model import DiffResponseSchema, split_into_reference_and_hash


@click.command("diff")
@click.argument("from_ref", nargs=1, required=True)
@click.argument("to_ref", nargs=1, required=True)
@pass_client
@error_handler
def diff(ctx: ContextObject, from_ref: str, to_ref: str) -> None:  # noqa: C901
    """Show diff between two given references.

    'from_ref'/'to_ref': name of branch or tag to use to show the diff

    'from_ref' and 'to_ref' can be either:

        The name of a reference (branch or tag), resolving to the HEAD of the
        named reference.

        The name of a reference (branch or tag) with a commit-id on that named
        reference.

        A detached commit-id/hash. Example:

    Examples:

        nessie diff from_ref to_ref -> show diff between 'from_ref' and 'to_ref', using the HEADs of both references

        nessie diff main@1234567890abcdef my-branch -> compare main branch at commit-id 1234567890abcdef with the HEAD of my-branch

        nessie diff 1234567890abcdef main -> compare the main branch w/ commit-id 1234567890abcdef

    """
    from_ref, from_hash_on_ref = split_into_reference_and_hash(from_ref)

    to_ref, to_hash_on_ref = split_into_reference_and_hash(to_ref)

    diff_response = ctx.nessie.get_diff(from_ref=from_ref, to_ref=to_ref, from_hash_on_ref=from_hash_on_ref, to_hash_on_ref=to_hash_on_ref)
    if ctx.json:
        click.echo(DiffResponseSchema().dumps(diff_response))
    else:
        click.echo_via_pager(x.pretty_print() + "\n" for index, x in enumerate(diff_response.diffs))



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Main module."""

from typing import Any, Generator, Optional, cast

import confuse

from pynessie.auth import setup_auth
from pynessie.client._endpoints import (
    all_references,
    assign_branch,
    assign_tag,
    cherry_pick,
    commit,
    create_reference,
    delete_branch,
    delete_tag,
    get_content,
    get_default_branch,
    get_diff,
    get_reference,
    list_logs,
    list_reflog,
    list_tables,
    merge,
)
from pynessie.error import NessieInvalidUsageException
from pynessie.model import (
    DETACHED_REFERENCE_NAME,
    Branch,
    CommitMeta,
    Content,
    ContentKey,
    ContentSchema,
    Detached,
    DiffResponse,
    DiffResponseSchema,
    Entries,
    EntriesSchema,
    LogEntry,
    LogResponse,
    LogResponseSchema,
    Merge,
    MergeResponse,
    MergeResponseSchema,
    MergeSchema,
    MultiContents,
    MultiContentSchema,
    Operation,
    Reference,
    ReferenceSchema,
    ReferencesResponse,
    ReferencesResponseSchema,
    ReflogEntry,
    ReflogResponse,
    ReflogResponseSchema,
    Tag,
    Transplant,
    TransplantSchema,
    split_into_reference_and_hash,
)


class NessieClient:
    """Base Nessie Client."""

    def __init__(self, config: confuse.Configuration) -> None:
        """Create a Nessie Client from known config."""
        self._base_url = config["endpoint"].get()
        self._ssl_verify = config["verify"].get(bool)
        self._auth = setup_auth(config)
        self._commit_id: str = cast(str, None)

        try:
            self._base_branch = config["default_branch"].get()
        except confuse.exceptions.NotFoundError:
            self._base_branch = None

    def list_references(self, fetch_all: bool = False) -> ReferencesResponse:
        """Fetch all known references.

        :return: list of Nessie References
        """
        references = all_references(self._base_url, self._auth, self._ssl_verify, fetch_all)
        return ReferencesResponseSchema().load(references)

    def get_reference(self, name: Optional[str]) -> Reference:
        """Fetch a ref.

        :param name: name of ref to fetch
        :return: Nessie reference
        """
        ref_obj = (
            get_reference(self._base_url, self._auth, name, self._ssl_verify)
            if name
            else get_default_branch(self._base_url, self._auth, self._ssl_verify)
        )
        ref = ReferenceSchema().load(ref_obj)
        return ref

    def create_branch(self, branch: str, ref: Optional[str] = None, hash_on_ref: Optional[str] = None) -> Branch:
        """Create a branch.

        :param branch: name of new branch to create
        :param ref: branch or tag to fork from ('hash_on_ref' should be reachable from this reference)
        :param hash_on_ref: hash to assign 'branch' to
        :return: Nessie branch object
        """
        ref_json = ReferenceSchema().dump(Branch(branch, hash_on_ref))
        ref_obj = create_reference(self._base_url, self._auth, ref_json, ref, self._ssl_verify)
        return cast(Branch, ReferenceSchema().load(ref_obj))

    def delete_branch(self, branch: str, hash_: str) -> None:
        """Delete a branch.

        :param branch: name of branch to delete
        :param hash_: hash of the branch
        """
        delete_branch(self._base_url, self._auth, branch, hash_, self._ssl_verify)

    def create_tag(self, tag: str, ref: str, hash_on_ref: Optional[str] = None) -> Tag:
        """Create a tag.

        :param tag: name of new tag to create
        :param ref: name of the reference via which 'hash_on_ref' is reachable
        :param hash_on_ref: hash to assign 'branch' to
        :return: Nessie tag object
        """
        ref_json = ReferenceSchema().dump(Tag(tag, hash_on_ref) if hash_on_ref else Tag(tag))
        ref_obj = create_reference(self._base_url, self._auth, ref_json, ref, self._ssl_verify)
        return cast(Tag, ReferenceSchema().load(ref_obj))

    def delete_tag(self, tag: str, hash_: str) -> None:
        """Delete a tag.

        :param tag: name of tag to delete
        :param hash_: hash of the branch
        """
        delete_tag(self._base_url, self._auth, tag, hash_, self._ssl_verify)

    def list_keys(
        self,
        ref: str,
        hash_on_ref: Optional[str] = None,
        max_result_hint: Optional[int] = None,
        page_token: Optional[str] = None,
        query_filter: Optional[str] = None,
    ) -> Entries:
        """Fetch a list of all tables from a known branch.

        :param ref: name of branch
        :param hash_on_ref: hash on reference
        :param entity_types: list of types to filter keys on
        :param query_filter: A CEL expression that allows advanced filtering capabilities
        :return: list of Nessie table names
        """
        ref_name, ref_hash = split_into_reference_and_hash(ref)
        if hash_on_ref:
            if ref_hash and ref_hash != hash_on_ref:
                raise NessieInvalidUsageException(
                    "Must not specify hash-on-ref using 'name@hash' and explicit hash-on-ref argument, use only one of those"
                )
        else:
            hash_on_ref = ref_hash

        return EntriesSchema().load(
            list_tables(self._base_url, self._auth, ref_name, hash_on_ref, max_result_hint, page_token, query_filter, self._ssl_verify)
        )

    def get_content(self, ref: str, content_key: ContentKey, hash_on_ref: Optional[str] = None) -> Content:
        """Fetch a content from a known ref.

        :param ref: name of ref
        :param hash_on_ref: hash on reference
        :param content_key: content key to fetch
        :return: A single content
        """
        ref_name, ref_hash = split_into_reference_and_hash(ref)
        if hash_on_ref:
            if ref_hash and ref_hash != hash_on_ref:
                raise NessieInvalidUsageException(
                    "Must not specify hash-on-ref using 'name@hash' and explicit hash-on-ref argument, use only one of those"
                )
        else:
            hash_on_ref = ref_hash

        return ContentSchema().load(get_content(self._base_url, self._auth, ref_name, content_key, hash_on_ref, self._ssl_verify))

    # pylint: disable=keyword-arg-before-vararg
    def commit(self, branch: str, old_hash: str, reason: Optional[str] = None, author: Optional[str] = None, *ops: Operation) -> Branch:
        """Modify a set of Nessie tables."""
        meta = CommitMeta(message=reason if reason else "")
        if author:
            meta.author = author
        ref_obj = commit(self._base_url, self._auth, branch, MultiContentSchema().dumps(MultiContents(meta, list(ops))), old_hash)
        return cast(Branch, ReferenceSchema().load(ref_obj))

    def _assign_to(self, to_ref: str, to_ref_hash: Optional[str] = None) -> Reference:
        ref_name, ref_hash = split_into_reference_and_hash(to_ref)

        if ref_hash and to_ref_hash and ref_hash != to_ref_hash:
            raise NessieInvalidUsageException(
                "Must not specify hash on to-ref using 'name@hash' and via explicit to-ref-hash argument, use only one of those"
            )

        if not ref_hash:
            ref_hash = to_ref_hash

        if ref_name == DETACHED_REFERENCE_NAME:
            return Detached(DETACHED_REFERENCE_NAME, ref_hash)

        ref = self.get_reference(ref_name)
        if ref_hash:
            ref.hash_ = ref_hash

        return ref

    def assign_branch(self, branch: str, to_ref: str, to_ref_hash: Optional[str] = None, old_hash: Optional[str] = None) -> None:
        """Assign a hash to a branch."""
        if not old_hash:
            old_hash = self.get_reference(branch).hash_
        assert old_hash is not None
        ref_json = ReferenceSchema().dumps(self._assign_to(to_ref, to_ref_hash))
        assign_branch(self._base_url, self._auth, branch, ref_json, old_hash, self._ssl_verify)

    def assign_tag(self, tag: str, to_ref: str, to_ref_hash: Optional[str] = None, old_hash: Optional[str] = None) -> None:
        """Assign a hash to a tag."""
        if not old_hash:
            old_hash = self.get_reference(tag).hash_
        assert old_hash is not None
        ref_json = ReferenceSchema().dumps(self._assign_to(to_ref, to_ref_hash))
        assign_tag(self._base_url, self._auth, tag, ref_json, old_hash, self._ssl_verify)

    def merge(
        self, from_ref: str, onto_branch: str, from_hash: Optional[str] = None, old_hash: Optional[str] = None
    ) -> Optional[MergeResponse]:
        """Merge a branch into another branch.

        Note:
            For legacy servers that return no response, method returns None.
            For servers running nessie-quarkus > 0.30.0, a None response would imply error.
        """
        onto_branch, old_hash_ref = split_into_reference_and_hash(onto_branch)
        if not old_hash:
            old_hash = old_hash_ref if old_hash_ref else self.get_reference(onto_branch).hash_
        elif old_hash_ref:
            if old_hash_ref and old_hash_ref != old_hash:
                raise NessieInvalidUsageException(
                    "Must not specify hash on from-ref using 'name@hash' and via explicit hash argument, use only one of those"
                )
            old_hash = old_hash_ref
        assert old_hash is not None

        from_ref, from_hash_ref = split_into_reference_and_hash(from_ref)
        if not from_hash:
            from_hash = from_hash_ref if from_hash_ref else self.get_reference(from_ref).hash_
        elif from_hash_ref:
            if from_hash_ref and from_hash_ref != from_hash:
                raise NessieInvalidUsageException(
                    "Must not specify hash on from-ref using 'name@hash' and via explicit hash argument, use only one of those"
                )
            from_hash = from_hash_ref

        merge_json = MergeSchema().dump(Merge(from_ref, str(from_hash)))
        merge_response = merge(self._base_url, self._auth, onto_branch, merge_json, old_hash, self._ssl_verify)
        if merge_response:
            return MergeResponseSchema().load(merge_response)
        return None

    # pylint: disable=keyword-arg-before-vararg
    def cherry_pick(self, branch: str, from_ref: str, old_hash: Optional[str] = None, *hashes: str) -> None:
        """Cherry pick a list of hashes to a branch."""
        if not old_hash:
            old_hash = self.get_reference(branch).hash_
        assert old_hash is not None
        transplant_json = TransplantSchema().dump(Transplant(from_ref, list(hashes)))
        cherry_pick(self._base_url, self._auth, branch, transplant_json, old_hash, self._ssl_verify)

    def get_log(
        self,
        start_ref: str,
        hash_on_ref: Optional[str] = None,
        max_records: Optional[int] = None,
        fetch_all: bool = False,
        **filtering_args: Any
    ) -> Generator[LogEntry, Any, None]:
        """Fetch all logs starting at start_ref.

        start_ref can be any ref.

        Note:
            this will load the log into local memory and filter at the client. Currently there are no
            primitives in the REST api to limit logs or perform paging. TODO
        """
        page_token = filtering_args.get("pageToken", None)

        def fetch_logs(fetch_max: Optional[int], token: Optional[str] = page_token) -> LogResponse:
            if token:
                filtering_args["pageToken"] = token

            fetched_logs = list_logs(
                base_url=self._base_url,
                auth=self._auth,
                hash_on_ref=hash_on_ref,
                ref=start_ref,
                ssl_verify=self._ssl_verify,
                max_records=fetch_max,
                fetch_all=fetch_all,
                **filtering_args
            )
            parsed_logs = LogResponseSchema().load(fetched_logs)
            return parsed_logs

        log_response = fetch_logs(fetch_max=max_records)

        def generator(logs: LogResponse, fetch_max: Optional[int]) -> Generator[LogEntry, Any, None]:
            while True:
                for log in logs.log_entries:
                    yield log
                    if fetch_max is not None:
                        fetch_max -= 1
                        if fetch_max <= 0:
                            # yield only the required number of results, if server returns more records than expected.
                            break
                if not logs.has_more or (fetch_max is not None and fetch_max <= 0):
                    break
                logs = fetch_logs(fetch_max=fetch_max, token=logs.token)

        return generator(log_response, max_records)

    def get_default_branch(self) -> str:
        """Fetch default branch either from config if specified or from the server."""
        return self._base_branch if self._base_branch else self.get_reference(None).name

    def get_base_url(self) -> str:
        """Return Nessie server configured base URL."""
        return self._base_url

    def get_diff(
        self, from_ref: str, to_ref: str, from_hash_on_ref: Optional[str] = None, to_hash_on_ref: Optional[str] = None
    ) -> DiffResponse:
        """Retrieve the diff between from_ref and to_ref.

        from_ref / to_ref can be any ref.
        """
        return DiffResponseSchema().load(
            get_diff(self._base_url, self._auth, from_ref, to_ref, from_hash_on_ref, to_hash_on_ref, self._ssl_verify)
        )

    def get_reflog(self, max_records: Optional[int] = None, **query_params: Any) -> Generator[ReflogEntry, Any, None]:
        """Fetch all reflog starting from the head or from the specified range."""
        page_token = query_params.get("pageToken", None)

        def fetch_reflog(fetch_max: Optional[int], token: Optional[str] = page_token) -> ReflogResponse:
            if token:
                query_params["pageToken"] = token

            fetched_reflog = list_reflog(
                base_url=self._base_url, auth=self._auth, ssl_verify=self._ssl_verify, max_records=fetch_max, **query_params
            )
            parsed_logs = ReflogResponseSchema().load(fetched_reflog)
            return parsed_logs

        reflog_response = fetch_reflog(fetch_max=max_records)

        def generator(logs: ReflogResponse, fetch_max: Optional[int]) -> Generator[ReflogEntry, Any, None]:
            while True:
                for log in logs.log_entries:
                    yield log
                    if fetch_max is not None:
                        fetch_max -= 1
                        if fetch_max <= 0:
                            # yield only the required number of results, if server returns more records than expected.
                            break
                if not logs.has_more or (fetch_max is not None and fetch_max <= 0):
                    break
                logs = fetch_reflog(fetch_max=fetch_max, token=logs.token)

        return generator(reflog_response, max_records)



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Configure pytest."""

import os
import shutil
import tempfile
from typing import Any, List, Optional

import attr
import pytest
import simplejson
from assertpy import assert_that
from click.testing import CliRunner, Result
from vcr.config import VCR
from vcr.request import Request
from vcr.serializers import yamlserializer

from pynessie import cli
from pynessie.model import Content, ContentSchema, ReferenceSchema


@attr.dataclass
class NessieTestConfig:
    """Test configs for pynessie tests."""

    config_dir: str
    cleanup: bool


nessie_test_config: NessieTestConfig = NessieTestConfig("", False)


def pytest_configure(config):  # noqa
    """Configure pytest."""
    config.addinivalue_line("markers", "doc: mark as end-to-end test.")


def pytest_sessionstart(session: pytest.Session) -> None:
    """Setup a fresh temporary config directory for tests."""
    nessie_test_config.config_dir = tempfile.mkdtemp() + "/"
    # Instruct Confuse to keep Nessie config file in the temp location:
    os.environ["NESSIEDIR"] = nessie_test_config.config_dir


def pytest_sessionfinish(session: pytest.Session, exitstatus: int) -> None:
    """Remove temporary config directory."""
    shutil.rmtree(nessie_test_config.config_dir)


def execute_cli_command_raw(args: List[str], input_data: Optional[str] = None, ret_val: int = 0) -> Result:
    """Execute a Nessie CLI command."""
    result = CliRunner().invoke(cli.cli, args, input=input_data)
    if result.exit_code != ret_val:
        print(result.stdout)
        print(result.stderr)
        print(result)  # exception
    assert_that(result.exit_code).is_equal_to(ret_val)
    return result


def execute_cli_command(args: List[str], input_data: Optional[str] = None, ret_val: int = 0) -> str:
    """Execute a Nessie CLI command and return its STDOUT."""
    return execute_cli_command_raw(args, input_data=input_data, ret_val=ret_val).stdout


def ref_hash(ref: str) -> str:
    """Get the hash for a reference."""
    refs = ReferenceSchema().loads(execute_cli_command(["--json", "branch", "-l"]), many=True)
    return next(i.hash_ for i in refs if i.name == ref)


def make_commit(
    key: str, table: Content, branch: str, head_hash: Optional[str] = None, message: str = "test message", author: str = "nessie test"
) -> None:
    """Make commit through Nessie CLI."""
    if not head_hash:
        refs = {i.name: i.hash_ for i in ReferenceSchema().loads(execute_cli_command(["--json", "branch"]), many=True)}
        head_hash = refs[branch]
    assert isinstance(head_hash, str)
    execute_cli_command(
        ["content", "commit", "--stdin", key, "--ref", branch, "-m", message, "-c", head_hash, "--author", author],
        input_data=ContentSchema().dumps(table),
    )


def reset_nessie_server_state() -> None:
    """Resets the Nessie Server to an initial, clean state for testing."""
    # Delete all branches except main
    branches = ReferenceSchema().loads(execute_cli_command(["--json", "branch"]), many=True)
    for branch in branches:
        if branch.name != "main":
            execute_cli_command(["branch", "-d", branch.name])

    # Delete all tags
    tags = ReferenceSchema().loads(execute_cli_command(["--json", "tag"]), many=True)
    for tag in tags:
        execute_cli_command(["tag", "-d", tag.name])

    # Note: This hash should match the java constant AbstractDatabaseAdapter.NO_ANCESTOR
    no_ancestor_hash = "2e1cfa82b035c26cbbbdae632cea070514eb8b773f616aaeaf668e2f0be8f10d"

    # Reset the main branch to the "root" (a.k.a. no ancestor) hash
    execute_cli_command(["branch", "--force", "-o", no_ancestor_hash, "main", "main"])

    # Verify that the main branch has been reset
    branches = ReferenceSchema().loads(execute_cli_command(["--json", "branch"]), many=True)
    assert_that(branches).is_length(1)
    assert_that(branches[0].name).is_equal_to("main")
    assert_that(branches[0].hash_).is_equal_to(no_ancestor_hash)


def before_record_cb(request: Request) -> Optional[Request]:
    """VCR callback that instructs it to not record our "cleanup" requests."""
    if nessie_test_config.cleanup:
        return None
    return request


def _sort_json(value: str) -> str:
    """Parses JSON and serializes it again with a fixed key order."""
    try:
        json = simplejson.loads(value)
        sorted_json = simplejson.dumps(json, sort_keys=True)
        return sorted_json
    except ValueError:
        return value


def _reformat_body(obj: dict) -> None:
    """Sorts JSON objects embedded in HTTP payload by their keys."""
    body = obj["body"]
    sorted_json = None
    if body:
        if isinstance(body, str):
            sorted_json = _sort_json(body)
            obj["body"] = sorted_json
        elif isinstance(body, dict):
            sorted_json = _sort_json(body["string"])
            body["string"] = sorted_json
        else:
            raise RuntimeError("Unexpected cassette structure")

    # Adjust content length if changed
    if sorted_json:
        headers = obj["headers"]
        for k in headers:
            if k.lower() == "content-length":  # might have different case in cassettes
                headers[k] = [str(len(sorted_json))]  # one-element list
                break


class _NessieSortingSerializer:
    @staticmethod
    def deserialize(cassette_string: str) -> Any:
        """Delegates to the default yaml cassette serializer."""
        return yamlserializer.deserialize(cassette_string)

    @staticmethod
    def serialize(cassette_dict: dict) -> str:
        """Reformats JSON payloads in requests and responses.

        This is to avoid spurious changes in cassette yaml files under source control.
        """
        if "interactions" in cassette_dict:
            for i in cassette_dict["interactions"]:
                _reformat_body(i["request"])
                _reformat_body(i["response"])

        return yamlserializer.serialize(cassette_dict)


def pytest_recording_configure(config: pytest.Config, vcr: VCR) -> None:
    """Callback invoked by pytest_recording. We register a custom VCR serializer here."""
    ser_name = "nessie_sorted_json"
    vcr.register_serializer(ser_name, _NessieSortingSerializer())
    vcr.serializer = ser_name


@pytest.fixture(scope="module")
def vcr_config() -> dict:
    """VCR config that adds a custom before_record_request callback."""
    nessie_test_config.cleanup = False
    return {
        "before_record_request": before_record_cb,
    }


@pytest.fixture(autouse=True)
def _clean_nessie_session_marker(request: pytest.FixtureRequest, record_mode: str) -> None:
    """This pytest fixture is invoked for all test methods and cleans up the Nessie Server state.

    :param request Request object provided by the pytest framework
    :param record_mode Recording mode provided by the VCR plugin
    """
    # Cleaning the Nessie Server state is meaningful only when we are recording,
    # i.e. when the tests are running against a real server, not against VCR cassettes.
    # Consequently, when the tests will run against VCR cassettes, the pre-recorded
    # responses will simulate a "clean" server.
    if record_mode is None or record_mode == "none":
        return

    vcr_marker = request.node.get_closest_marker("vcr")

    # Clean the server state for tests that have the @pytest.mark.vcr annotation
    if vcr_marker:
        if "record_mode" in vcr_marker.kwargs:
            # if we have record_mode set in @pytest.mark.vcr annotation, we test that first
            test_record_mode = vcr_marker.kwargs["record_mode"]

            if test_record_mode is None:
                raise RuntimeError("record_mode can't be None or empty if set.")

            if test_record_mode == "none":
                return

        try:
            nessie_test_config.cleanup = True
            reset_nessie_server_state()
        finally:
            nessie_test_config.cleanup = False



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Tests for `pynessie` package."""
import itertools
from typing import List

import confuse
import pytest
import simplejson
from assertpy import assert_that

from pynessie import __version__
from pynessie.model import (
    Branch,
    CommitMeta,
    CommitMetaSchema,
    ContentKey,
    ContentSchema,
    DiffResponseSchema,
    EntrySchema,
    IcebergTable,
    LogEntry,
    LogEntrySchema,
    MergeResponseSchema,
    ReferenceSchema,
    ReflogEntry,
    ReflogEntrySchema,
)

from .conftest import (
    execute_cli_command,
    execute_cli_command_raw,
    make_commit,
    ref_hash,
)


@pytest.mark.vcr
def test_command_line_interface() -> None:
    """Test the CLI."""
    assert "Usage: nessie" in execute_cli_command([])
    assert "Usage: nessie" in execute_cli_command(["--help"])
    assert __version__ in execute_cli_command(["--version"])
    references = ReferenceSchema().loads(execute_cli_command(["--json", "branch", "-l"]), many=True)
    assert len(references) == 1
    assert references[0].name == "main"
    assert isinstance(references[0], Branch)


def test_config_options() -> None:
    """Ensure config cli option is consistent."""
    assert "Usage: nessie" in execute_cli_command(["config"])
    vars_to_add = ["--add x", "--get x", "--list", "--unset x"]
    for i in itertools.permutations(vars_to_add, 2):
        assert "Error: Illegal usage: " in execute_cli_command(["config"] + [*i[0].split(" "), *i[1].split(" ")], ret_val=2)

    execute_cli_command(["config", "x", "--add", "x"])


def test_set_unset() -> None:
    """Test config set/unset/list."""
    execute_cli_command(["config", "--add", "test.data", "123", "--type", "int"])
    assert execute_cli_command(["config", "test.data", "--type", "int"]) == "123\n"
    execute_cli_command(["config", "--unset", "test.data"])
    assert "123" not in execute_cli_command(["config", "--list"])


@pytest.mark.vcr
def test_remote() -> None:
    """Test setting and viewing remote."""
    execute_cli_command(["remote", "add", "http://test.url"])
    execute_cli_command(["remote", "add", "http://localhost:19120/api/v1"])
    assert "main" in execute_cli_command(["--json", "remote", "show"])
    execute_cli_command(["remote", "set-head", "dev"])
    assert execute_cli_command(["config", "default_branch"]) == "dev\n"
    execute_cli_command(["remote", "set-head", "dev", "-d"])
    result = execute_cli_command_raw(["config", "default_branch"], ret_val=1)
    assert result.stdout == ""
    assert isinstance(result.exception, confuse.exceptions.ConfigTypeError)


def _new_table(table_id: str) -> IcebergTable:
    return IcebergTable(table_id, "/a/b/c", 42, 43, 44, 45)


@pytest.mark.vcr
def test_log() -> None:
    """Test log and log filtering."""
    main_hash = ref_hash("main")

    logs = simplejson.loads(execute_cli_command(["--json", "log"]))
    assert len(logs) == 0
    execute_cli_command(["branch", "dev_test_log"])
    table = _new_table("test_log_dev")
    make_commit("log.foo.dev", table, "dev_test_log", author="nessie_user1")
    table = _new_table("test_log")
    make_commit("log.foo.bar", table, "main", author="nessie_user1", message="commit to main")
    tables = ContentSchema().loads(execute_cli_command(["--json", "content", "view", "log.foo.bar"]), many=True)
    assert len(tables) == 1
    assert tables[0] == table

    ext_logs: List[LogEntry] = LogEntrySchema().loads(execute_cli_command(["--json", "log", "-x"]), many=True)
    assert (
        len(ext_logs) == 1
        and ext_logs[0].commit_meta.message == "commit to main"
        and ext_logs[0].commit_meta.author == "nessie_user1"
        and ext_logs[0].parent_commit_hash is not None
        and len(ext_logs[0].operations) == 1
        and ext_logs[0].operations[0].key == ContentKey.from_path_string("log.foo.bar")
    )

    simple_logs: List[CommitMeta] = CommitMetaSchema().loads(execute_cli_command(["--json", "log"]), many=True)
    assert len(simple_logs) == 1 and simple_logs[0].message == "commit to main" and simple_logs[0].author == "nessie_user1"

    logs_by_hash: List[CommitMeta] = CommitMetaSchema().loads(execute_cli_command(["--json", "log", simple_logs[0].hash_]), many=True)
    assert_that(logs_by_hash).is_equal_to(simple_logs)

    logs = simplejson.loads(execute_cli_command(["--json", "log"]))
    logs_hash = logs[0]["hash"]
    assert_that(logs).is_length(1)
    assert_that(simplejson.loads(execute_cli_command(["--json", "log", "--revision-range", logs_hash]))).is_equal_to(logs)
    assert_that(simplejson.loads(execute_cli_command(["--json", "log", "--revision-range", f"..{logs_hash}"]))).is_equal_to(logs)
    assert_that(simplejson.loads(execute_cli_command(["--json", "log", f"main@{logs_hash}", "--revision-range", logs_hash]))).is_equal_to(
        logs
    )

    entries = EntrySchema().loads(execute_cli_command(["--json", "content", "list"]), many=True)
    assert_that(entries).is_length(1)
    execute_cli_command(
        [
            "--json",
            "content",
            "commit",
            "log.foo.bar",
            "-R",
            "--ref",
            "main",
            "-m",
            "delete_message",
            "-c",
            logs[0]["hash"],
            "--author",
            "nessie_user2",
        ],
    )
    assert_that(simplejson.loads(execute_cli_command(["--json", "log", "-n", "1"]))).is_length(1)
    assert_that(simplejson.loads(execute_cli_command(["--json", "log", "dev_test_log"]))).is_length(1)
    logs = simplejson.loads(execute_cli_command(["--json", "log"]))
    assert_that(logs).is_length(2)
    assert_that(
        simplejson.loads(execute_cli_command(["--json", "log", "--revision-range", "{}..{}".format(logs[0]["hash"], logs[1]["hash"])]))
    ).is_length(1)
    assert_that(simplejson.loads(execute_cli_command(["--json", "log"]))).is_length(2)
    logs = simplejson.loads(execute_cli_command(["--json", "log", "--author", "nessie_user1"]))
    assert_that(logs).is_length(1)
    assert_that(logs[0]["author"]).is_equal_to("nessie_user1")
    logs = simplejson.loads(execute_cli_command(["--json", "log", "--author", "nessie_user2"]))
    assert_that(logs).is_length(1)
    assert_that(logs[0]["author"]).is_equal_to("nessie_user2")
    assert_that(simplejson.loads(execute_cli_command(["--json", "log", "--author", "nessie_user2", "--author", "nessie_user1"]))).is_length(
        2
    )
    # the committer is set on the server-side and is empty if we're not logged
    # in when performing a commit
    assert_that(simplejson.loads(execute_cli_command(["--json", "log", "--committer", ""]))).is_length(2)
    assert_that(
        simplejson.loads(
            execute_cli_command(["--json", "log", "--filter", "commit.author == 'nessie_user2' || commit.author == 'non_existing'"])
        )
    ).is_length(1)
    assert_that(
        simplejson.loads(
            execute_cli_command(["--json", "log", "--after", "2001-01-01T00:00:00+00:00", "--before", "2999-12-30T23:00:00+00:00"])
        )
    ).is_length(2)

    # Specifying a different end-hash via revision-range and ref is forbidden
    execute_cli_command(["--json", "log", main_hash, "--revision-range", logs[0]["hash"]], ret_val=2)
    execute_cli_command(["--json", "log", f"main@{main_hash}", "--revision-range", logs[0]["hash"]], ret_val=2)


@pytest.mark.vcr
def test_branch() -> None:
    """Test create and assign refs."""
    main_hash = ref_hash("main")

    references = ReferenceSchema().loads(execute_cli_command(["--json", "branch"]), many=True)
    assert len(references) == 1
    execute_cli_command(["branch", "dev"])
    references = ReferenceSchema().loads(execute_cli_command(["--json", "branch"]), many=True)
    assert len(references) == 2
    execute_cli_command(["branch", "etl", "main"])
    references = ReferenceSchema().loads(execute_cli_command(["--json", "branch"]), many=True)
    assert len(references) == 3
    execute_cli_command(["branch", "dev_hash", main_hash])
    references = ReferenceSchema().loads(execute_cli_command(["--json", "branch"]), many=True)
    assert len(references) == 4
    execute_cli_command(["branch", "etl_hash", f"main@{main_hash}"])
    references = ReferenceSchema().loads(execute_cli_command(["--json", "branch"]), many=True)
    assert len(references) == 5
    references = ReferenceSchema().loads(execute_cli_command(["--json", "branch", "-l", "etl"]), many=False)
    assert_that(references.name).is_equal_to("etl")
    references = simplejson.loads(execute_cli_command(["--json", "branch", "-l", "foo"]))
    assert len(references) == 0

    table = _new_table("test_branch_metadata")
    make_commit("test.branch.metadata", table, "dev", author="nessie_user1")

    branch = ReferenceSchema().loads(execute_cli_command(["--json", "branch", "-l", "dev", "--extended"]))
    ref_metadata = branch.metadata
    assert_that(ref_metadata).is_not_none()
    assert_that(ref_metadata.num_commits_ahead).is_equal_to(1)
    assert_that(ref_metadata.num_commits_behind).is_equal_to(0)
    assert_that(ref_metadata.num_total_commits).is_equal_to(1)
    assert_that(ref_metadata.common_ancestor_hash).is_not_empty()
    assert_that(ref_metadata.commit_meta_of_head).is_not_none()

    dev_hash = ref_hash("dev")

    execute_cli_command(["branch", "-d", "etl"])
    execute_cli_command(["branch", "-d", "dev", "-c", dev_hash])
    execute_cli_command(["branch", "-d", "etl_hash", "-c", main_hash])
    execute_cli_command(["branch", "-d", "dev_hash"])
    references = ReferenceSchema().loads(execute_cli_command(["--json", "branch"]), many=True)
    assert len(references) == 1


@pytest.mark.vcr
def test_tag() -> None:
    """Test create and assign refs."""
    main_hash = ref_hash("main")

    references = ReferenceSchema().loads(execute_cli_command(["--json", "tag"]), many=True)
    assert len(references) == 0
    execute_cli_command(["tag", "dev-tag", "main"])
    references = ReferenceSchema().loads(execute_cli_command(["--json", "tag"]), many=True)
    assert len(references) == 1
    execute_cli_command(["tag", "etl-tag", "main"])
    references = ReferenceSchema().loads(execute_cli_command(["--json", "tag"]), many=True)
    assert len(references) == 2
    execute_cli_command(["tag", "dev-hash-tag", main_hash])
    references = ReferenceSchema().loads(execute_cli_command(["--json", "tag"]), many=True)
    assert len(references) == 3
    execute_cli_command(["tag", "etl-hash-tag", f"main@{main_hash}"])
    references = ReferenceSchema().loads(execute_cli_command(["--json", "tag"]), many=True)
    assert len(references) == 4
    references = ReferenceSchema().loads(execute_cli_command(["--json", "tag", "-l", "etl-tag"]), many=False)
    assert_that(references.name).is_equal_to("etl-tag")
    references = simplejson.loads(execute_cli_command(["--json", "tag", "-l", "foo"]))
    assert len(references) == 0
    execute_cli_command(["tag", "-d", "etl-tag"])
    execute_cli_command(["tag", "-d", "etl-hash-tag"])
    execute_cli_command(["tag", "-d", "dev-tag", "-c", main_hash])
    execute_cli_command(["tag", "-d", "dev-hash-tag", "-c", main_hash])
    references = ReferenceSchema().loads(execute_cli_command(["--json", "tag"]), many=True)
    assert len(references) == 0
    execute_cli_command(["tag", "v1.0"])
    tags = {i.name: i.hash_ for i in ReferenceSchema().loads(execute_cli_command(["--json", "tag"]), many=True)}
    branches = {i.name: i.hash_ for i in ReferenceSchema().loads(execute_cli_command(["--json", "branch"]), many=True)}
    assert tags["v1.0"] == branches["main"]

    execute_cli_command(["branch", "metadata_branch", "main"])
    table = _new_table("test_tag_metadata")
    make_commit("test.tag.metadata", table, "metadata_branch", author="nessie_user1")
    execute_cli_command(["tag", "metadata_tag", "metadata_branch"])
    ref = ReferenceSchema().loads(execute_cli_command(["--json", "tag", "-l", "metadata_tag", "--extended"]))
    ref_metadata = ref.metadata
    assert_that(ref_metadata).is_not_none()
    assert_that(ref_metadata.num_commits_ahead).is_none()
    assert_that(ref_metadata.num_commits_behind).is_none()
    assert_that(ref_metadata.num_total_commits).is_equal_to(1)
    assert_that(ref_metadata.common_ancestor_hash).is_none()
    assert_that(ref_metadata.commit_meta_of_head).is_not_none()


@pytest.mark.vcr
def test_assign() -> None:
    """Test assign operation."""
    execute_cli_command(["branch", "dev"])
    make_commit("assign.foo.bar", _new_table("test_assign"), "dev")
    execute_cli_command(["branch", "main", "dev", "--force"])
    branches = ReferenceSchema().loads(execute_cli_command(["--json", "branch"]), many=True)
    refs = {i.name: i.hash_ for i in branches}
    assert refs["main"] == refs["dev"]
    execute_cli_command(["tag", "v1.0", "main"])
    tags = {i.name: i.hash_ for i in ReferenceSchema().loads(execute_cli_command(["--json", "tag"]), many=True)}
    assert tags["v1.0"] == refs["main"]
    execute_cli_command(["tag", "v1.0", "dev", "--force"])
    tags = {i.name: i.hash_ for i in ReferenceSchema().loads(execute_cli_command(["--json", "tag"]), many=True)}
    assert tags["v1.0"] == refs["dev"]


@pytest.mark.vcr
def test_merge() -> None:
    """Test merge operation."""
    execute_cli_command(["branch", "dev"])
    make_commit("merge.foo.bar", _new_table("test_merge"), "dev")
    main_hash = ref_hash("main")
    dev_hash = ref_hash("dev")

    # Passing detached commit-id plus a _different_ hash-on-ref --> error
    execute_cli_command(["merge", f"dev@{dev_hash}", "-c", main_hash, "-o", main_hash], ret_val=1)
    merge_output = execute_cli_command(["merge", "dev", "-c", main_hash])

    branches = ReferenceSchema().loads(execute_cli_command(["--json", "branch"]), many=True)
    refs = {i.name: i.hash_ for i in branches}

    expected_output_list = [
        "The following 1 commits were merged onto main:",
        '{} "test message" '.format(refs["dev"]),
        "Resultant hash on main after merge: {}".format(refs["main"]),
    ]
    assert_that(merge_output.splitlines()).is_equal_to(expected_output_list)

    # if we try to merge again from dev to main we get an error
    # this is because there is now a conflict as the same commit exists on both branches
    merge_error = execute_cli_command(["merge", "dev", "-c", refs["main"]], ret_val=1)
    assert "Client Error Conflict: The following keys have been changed in conflict:" in merge_error

    logs = simplejson.loads(execute_cli_command(["--json", "log"]))
    # we don't check for equality of hashes here because a merge
    # produces a different commit hash on the target branch
    assert len(logs) == 1
    assert_that(logs[0]["message"]).is_equal_to("test message")
    logs = simplejson.loads(execute_cli_command(["--json", "log", "dev"]))
    assert len(logs) == 1


@pytest.mark.vcr
def test_merge_json() -> None:
    """Test merge operation."""
    execute_cli_command(["branch", "dev"])
    make_commit("merge.foo.bar", _new_table("test_merge"), "dev")
    main_hash = ref_hash("main")
    dev_hash = ref_hash("dev")

    # Passing detached commit-id plus a _different_ hash-on-ref --> error
    execute_cli_command(["merge", f"dev@{dev_hash}", "-c", main_hash, "-o", main_hash], ret_val=1)

    merge_output = execute_cli_command(["--json", "merge", "dev", "-c", main_hash])
    merge_response = MergeResponseSchema().loads(merge_output)

    # Check merge response
    assert_that(merge_response.target_branch).is_equal_to("main")
    assert_that(merge_response.expected_hash).is_equal_to(main_hash)
    assert_that(merge_response.effective_target_hash).is_equal_to(main_hash)
    assert len(merge_response.source_commits) == 1
    assert len(merge_response.details) == 1
    assert_that(merge_response.details[0].source_commits[0]).is_equal_to(dev_hash)

    logs = simplejson.loads(execute_cli_command(["--json", "log"]))
    # we don't check for equality of hashes here because a merge
    # produces a different commit hash on the target branch
    assert len(logs) == 1
    assert_that(logs[0]["message"]).is_equal_to("test message")
    logs = simplejson.loads(execute_cli_command(["--json", "log", "dev"]))
    assert len(logs) == 1


@pytest.mark.vcr
def test_merge_legacy() -> None:
    """Test merge operation."""
    execute_cli_command(["branch", "dev"])
    make_commit("merge.foo.bar", _new_table("test_merge"), "dev")
    main_hash = ref_hash("main")

    merge_output = execute_cli_command(["merge", "dev", "-c", main_hash])
    assert_that(merge_output).is_equal_to("Merge succeeded but legacy server did not respond with additional details.\n")

    execute_cli_command(["branch", "devjson"])
    make_commit("merge.foo.far", _new_table("test_merge_json"), "devjson")

    merge_output = execute_cli_command(["--json", "merge", "-f", "-b", "main", "devjson"])
    assert_that(merge_output).is_equal_to('{"_cli_hint": "Merge succeeded but legacy server did not respond with additional details."}\n')


@pytest.mark.vcr
def test_merge_detached() -> None:
    """Test merge operation."""
    execute_cli_command(["branch", "dev"])
    make_commit("merge.foo.bar", _new_table("test_merge_detached"), "dev")
    main_hash = ref_hash("main")
    dev_hash = ref_hash("dev")

    # Passing detached commit-id plus a _different_ hash-on-ref --> error
    execute_cli_command(["merge", dev_hash, "-c", main_hash, "-o", main_hash], ret_val=1)

    merge_output = execute_cli_command(["merge", "dev", "-c", main_hash])

    branches = ReferenceSchema().loads(execute_cli_command(["--json", "branch"]), many=True)
    refs = {i.name: i.hash_ for i in branches}

    expected_output_list = [
        "The following 1 commits were merged onto main:",
        '{} "test message" '.format(refs["dev"]),
        "Resultant hash on main after merge: {}".format(refs["main"]),
    ]
    assert_that(merge_output.splitlines()).is_equal_to(expected_output_list)

    logs = simplejson.loads(execute_cli_command(["--json", "log"]))
    # we don't check for equality of hashes here because a merge
    # produces a different commit hash on the target branch
    assert len(logs) == 1
    assert_that(logs[0]["message"]).is_equal_to("test message")
    logs = simplejson.loads(execute_cli_command(["--json", "log", "dev"]))
    assert len(logs) == 1


@pytest.mark.vcr
def test_merge_detached_json() -> None:
    """Test merge operation."""
    execute_cli_command(["branch", "dev"])
    make_commit("merge.foo.bar", _new_table("test_merge_detached"), "dev")
    main_hash = ref_hash("main")
    dev_hash = ref_hash("dev")

    # Passing detached commit-id plus a _different_ hash-on-ref --> error
    execute_cli_command(["merge", dev_hash, "-c", main_hash, "-o", main_hash], ret_val=1)

    merge_output = execute_cli_command(["--json", "merge", dev_hash, "-c", main_hash])
    merge_response = MergeResponseSchema().loads(merge_output)

    # Check merge response
    assert_that(merge_response.target_branch).is_equal_to("main")
    assert_that(merge_response.expected_hash).is_equal_to(main_hash)
    assert_that(merge_response.effective_target_hash).is_equal_to(main_hash)
    assert len(merge_response.source_commits) == 1
    assert len(merge_response.details) == 1
    assert_that(merge_response.details[0].source_commits[0]).is_equal_to(dev_hash)

    logs = simplejson.loads(execute_cli_command(["--json", "log"]))
    # we don't check for equality of hashes here because a merge
    # produces a different commit hash on the target branch
    assert len(logs) == 1
    assert_that(logs[0]["message"]).is_equal_to("test message")
    logs = simplejson.loads(execute_cli_command(["--json", "log", "dev"]))
    assert len(logs) == 1


@pytest.mark.vcr
def test_transplant() -> None:
    """Test transplant operation."""
    execute_cli_command(["branch", "dev"])
    make_commit("transplant.foo.bar", _new_table("test_transplant_1"), "dev", message="commit 1")
    make_commit("bar.bar", _new_table("test_transplant_2"), "dev", message="commit 2")
    make_commit("foo.baz", _new_table("test_transplant_3"), "dev", message="commit 3")
    main_hash = ref_hash("main")
    logs = simplejson.loads(execute_cli_command(["--json", "log", "dev"]))
    first_hash = [i["hash"] for i in logs]
    execute_cli_command(["cherry-pick", "-c", main_hash, "-s", "dev", first_hash[1], first_hash[0]])

    logs = simplejson.loads(execute_cli_command(["--json", "log"]))
    assert len(logs) == 2  # two commits were transplanted into an empty `main`
    assert_that(logs[0]["message"]).is_equal_to("commit 3")
    assert_that(logs[1]["message"]).is_equal_to("commit 2")


@pytest.mark.vcr
def test_diff() -> None:
    """Test log and log filtering."""
    diff = DiffResponseSchema().loads(execute_cli_command(["--json", "diff", "main", "main"]))
    main_hash = ref_hash("main")
    assert_that(diff).is_not_none()
    assert_that(diff.diffs).is_empty()
    branch = "dev_test_diff"
    execute_cli_command(["branch", branch])
    table = _new_table(branch)
    content_key = "diff.foo.dev"
    make_commit(content_key, table, branch, author="nessie_user1")
    branch_hash = ref_hash(branch)

    diff = DiffResponseSchema().loads(execute_cli_command(["--json", "diff", "main", branch]))
    assert_that(diff).is_not_none()
    assert_that(diff.diffs).is_length(1)
    diff_entry = diff.diffs[0]
    assert_that(diff_entry.content_key).is_equal_to(ContentKey.from_path_string(content_key))
    assert_that(diff_entry.from_content).is_none()
    assert_that(diff_entry.to_content).is_equal_to(table)

    diff_detached = DiffResponseSchema().loads(execute_cli_command(["--json", "diff", main_hash, branch_hash]))
    assert_that(diff_detached).is_equal_to(diff)
    diff_detached = DiffResponseSchema().loads(execute_cli_command(["--json", "diff", f"DETACHED@{main_hash}", f"DETACHED@{branch_hash}"]))
    assert_that(diff_detached).is_equal_to(diff)
    diff_detached = DiffResponseSchema().loads(execute_cli_command(["--json", "diff", f"main@{main_hash}", f"{branch}@{branch_hash}"]))
    assert_that(diff_detached).is_equal_to(diff)


@pytest.mark.vcr
def test_reflog() -> None:
    """Test reflog."""
    expected_entries = []
    execute_cli_command(["branch", "dev_test_reflog1"])
    expected_entries.append(("dev_test_reflog1", "CREATE_REFERENCE"))
    #
    table = _new_table("dev_test_reflog1")
    content_key = "reflog.foo.dev"
    make_commit(content_key, table, "dev_test_reflog1", author="nessie_user1")
    expected_entries.append(("dev_test_reflog1", "COMMIT"))
    #
    execute_cli_command(["tag", "dev_test_reflog2"])
    expected_entries.append(("dev_test_reflog2", "CREATE_REFERENCE"))
    #
    execute_cli_command(["branch", "-d", "dev_test_reflog1"])
    expected_entries.append(("dev_test_reflog1", "DELETE_REFERENCE"))
    #
    execute_cli_command(["tag", "-d", "dev_test_reflog2"])
    expected_entries.append(("dev_test_reflog2", "DELETE_REFERENCE"))
    #
    response_entries: List[ReflogEntry] = ReflogEntrySchema().loads(execute_cli_command(["--json", "reflog"]), many=True)
    # latest entry in the reflog will be the head of reflog. Hence reverse the expected list
    expected_entries.reverse()
    # validate the entries
    assert_that(response_entries[0:5]).extracting("ref_name", "operation").is_equal_to(expected_entries)
    # test pagination
    response_entries_1: List[ReflogEntry] = ReflogEntrySchema().loads(execute_cli_command(["--json", "reflog", "-n", "2"]), many=True)
    assert_that(response_entries[0:2]).is_equal_to(response_entries_1)
    # test start hash (closer to the head)
    response_entries_2: List[ReflogEntry] = ReflogEntrySchema().loads(
        execute_cli_command(["--json", "reflog", "--revision-range", response_entries[2].reflog_id]), many=True
    )
    assert_that(response_entries[2:5]).is_equal_to(response_entries_2[0:3])
    # test range from 2 to 3 (both inclusive)
    response_entries_3: List[ReflogEntry] = ReflogEntrySchema().loads(
        execute_cli_command(
            ["--json", "reflog", "--revision-range", "{}..{}".format(response_entries[3].reflog_id, response_entries[2].reflog_id)]
        ),
        many=True,
    )
    assert_that(response_entries[2:4]).is_equal_to(response_entries_3[0:2])



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""merge CLI command."""

import json

import click
from click import UsageError

from pynessie.cli_common_context import ContextObject, MutuallyExclusiveOption
from pynessie.decorators import error_handler, pass_client, validate_reference
from pynessie.model import MergeResponseSchema


@click.command("merge")
@click.option("-b", "--branch", "ref", help="branch to merge onto. If not supplied the default branch from config is used")
@click.argument("from_ref", nargs=1, required=False)
@click.option(
    "-f",
    "--force",
    cls=MutuallyExclusiveOption,
    is_flag=True,
    mutually_exclusive=["condition"],
    help="force branch assignment",
)
@click.option(
    "-c",
    "--condition",
    "expected_hash",
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["force"],
    help="Expected hash. Only perform the action if the branch currently points to the hash specified by this option.",
)
@click.option("-o", "--hash-on-ref", help="Hash on merge-from-reference")
@pass_client
@error_handler
@validate_reference
def merge(ctx: ContextObject, ref: str, force: bool, expected_hash: str, hash_on_ref: str, from_ref: str) -> None:
    """Merge FROM_REF into another branch.

    FROM_REF can be a hash or branch.

    Examples:

        nessie merge -c 12345678abcdef dev -> merge dev to default branch with default branch's
    expected hash '12345678abcdef'

        nessie merge -b main -c 12345678abcdef dev -> merge dev to a branch named main with main branch's
    expected hash '12345678abcdef'

        nessie merge -b main -o 56781234abcdef -c 12345678abcdef dev -> merge dev at hash-on-ref '56781234abcdef' to
    main branch with main branch's expected hash '12345678abcdef'

        nessie merge -f -b main dev -> forcefully merge dev to a branch named main
    """
    if not force and not expected_hash:
        raise UsageError(
            """Either condition or force must be set. Condition should be set to a valid hash for concurrency
            control or force to ignore current state of Nessie Store."""
        )
    merge_response = ctx.nessie.merge(from_ref, ref, hash_on_ref, expected_hash)
    if merge_response is None:
        message = "Merge succeeded but legacy server did not respond with additional details."
        if ctx.json:
            hint = {"_cli_hint": message}
            click.echo(json.dumps(hint))
        else:
            click.echo(message)
    else:
        if ctx.json:
            click.echo(MergeResponseSchema().dumps(merge_response))
        else:
            commits = merge_response.source_commits
            click.echo(
                "The following {N} commits were merged onto {target_branch}:".format(
                    N=len(commits), target_branch=merge_response.target_branch
                )
            )
            for commit in commits:
                commit_meta = commit.commit_meta
                click.echo(commit_meta.hash_ + ' "' + commit_meta.message + '" ' + commit_meta.committer)
            if merge_response.resultant_target_hash:
                click.echo(
                    "Resultant hash on {} after merge: {}".format(merge_response.target_branch, merge_response.resultant_target_hash)
                )



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Tests for `config_parser.py`."""

from assertpy import assert_that

from pynessie.model import (
    DETACHED_REFERENCE_NAME,
    is_valid_hash,
    is_valid_reference_name,
    split_into_reference_and_hash,
)


def test_is_valid_reference_name() -> None:
    """Validates is_valid_reference_name."""
    assert_that(is_valid_reference_name("")).is_false()
    assert_that(is_valid_reference_name("*")).is_false()
    assert_that(is_valid_reference_name("@")).is_false()
    assert_that(is_valid_reference_name("main")).is_true()
    assert_that(is_valid_reference_name("abc/def")).is_true()


def test_is_valid_hash() -> None:
    """Validates is_valid_hash."""
    assert_that(is_valid_hash(".cafe")).is_false()
    assert_that(is_valid_hash("main")).is_false()
    assert_that(is_valid_hash("1122334455667788990011223344556677889900112233445566778899001122")).is_true()
    assert_that(is_valid_hash("abcDEF4242424242424242424242BEEF00DEAD42112233445566778899001122")).is_true()
    assert_that(is_valid_hash("caffee20")).is_true()
    assert_that(is_valid_hash("20caffee")).is_true()


def test_split_into_reference_and_hash() -> None:
    """Validates split_into_reference_and_hash."""
    assert_that(split_into_reference_and_hash("main")).is_equal_to(("main", None))
    assert_that(split_into_reference_and_hash("abcDEF4242424242424242424242BEEF00DEAD42112233445566778899001122")).is_equal_to(
        (DETACHED_REFERENCE_NAME, "abcDEF4242424242424242424242BEEF00DEAD42112233445566778899001122")
    )
    assert_that(split_into_reference_and_hash("main@caffee20")).is_equal_to(("main", "caffee20"))
    assert_that(split_into_reference_and_hash("main*caffee20")).is_equal_to(("main", "caffee20"))
    assert_that(split_into_reference_and_hash("DETACHED@1596af9580555a9a8437553878546385b070440acc063ac4461770ebce763aa1")).is_equal_to(
        (DETACHED_REFERENCE_NAME, "1596af9580555a9a8437553878546385b070440acc063ac4461770ebce763aa1")
    )
    assert_that(split_into_reference_and_hash("1596af9580555a9a8437553878546385b070440acc063ac4461770ebce763aa1")).is_equal_to(
        (DETACHED_REFERENCE_NAME, "1596af9580555a9a8437553878546385b070440acc063ac4461770ebce763aa1")
    )
    assert_that(split_into_reference_and_hash(None)).is_equal_to(("<UNKNOWN>", None))



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Top-level package for Nessie CLI commands."""

from pynessie.commands.branch import branch_
from pynessie.commands.cherry_pick import cherry_pick
from pynessie.commands.config import config
from pynessie.commands.content import content
from pynessie.commands.diff import diff
from pynessie.commands.log import log
from pynessie.commands.merge import merge
from pynessie.commands.reflog import reflog
from pynessie.commands.remote import remote
from pynessie.commands.tag import tag

__all__ = ["remote", "tag", "branch_", "cherry_pick", "config", "log", "merge", "content", "diff", "reflog"]



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Top-level decorators."""
import functools
import sys
from typing import Any, Callable

import click
from marshmallow import ValidationError

from pynessie.cli_common_context import ContextObject
from pynessie.error import NessieCliError, NessieException

# Decorator to pass Nessie client down into Click sub-commands.
pass_client = click.make_pass_decorator(ContextObject)


def validate_reference(f: Callable) -> Callable:
    """Decorator to validate Nessie reference and override it with default if needed."""

    @functools.wraps(f)
    def wrapper(*args: Any, **kwargs: Any) -> None:
        """Wrapper Decorator."""
        # Nessie context is always as the first argument
        if len(args) < 1 or not isinstance(args[0], ContextObject):
            raise RuntimeError("Use this decorator only with Nessie context.")

        ctx: ContextObject = args[0]

        # Check for ref and override with default if is not set
        if "ref" in kwargs:
            kwargs["ref"] = kwargs["ref"] if kwargs["ref"] else ctx.nessie.get_default_branch()

        return f(*args, **kwargs)

    return wrapper


def error_handler(f: Callable) -> Callable:
    """Decorator to catch and pretty print errors."""

    @functools.wraps(f)
    def wrapper(*args: Any, **kwargs: Any) -> None:
        """Wrapper object."""
        try:
            f(*args, **kwargs)
        except NessieCliError as e:
            if args[0].json:
                click.echo(e.json())
            else:
                click.echo(_format_cli_error(e.title, e.msg))
        except ValidationError as e:
            if args[0].json:
                click.echo(e.messages)
            else:
                click.echo(_format_cli_error("Validation Error", e.messages))
            sys.exit(1)
        except NessieException as e:
            if args[0].json:
                click.echo(e.json())
            else:
                click.echo(_format_error(e))
            sys.exit(1)

    return wrapper


def _format_error(e: NessieException) -> str:
    fmt = "{} (Status code: {})\n".format(click.style(e, fg="red"), e.status_code)
    fmt += "{} {}\n".format(click.style("Requested URL:", fg="yellow"), e.url)
    fmt += "{} {}\n".format(click.style("Server status:", fg="yellow"), e.server_status)
    fmt += "{} {}\n".format(click.style("Error code:", fg="yellow"), e.error_code)
    fmt += "{} {}\n".format(click.style("Server message:", fg="yellow"), e.server_message)
    fmt += "{} {}\n".format(click.style("Server traceback:", fg="yellow"), e.server_stack_trace)
    return fmt


def _format_cli_error(title: str, message: Any) -> str:
    fmt = "{}: \n".format(click.style(title, fg="red"))
    fmt += click.style(message, fg="yellow")
    return fmt



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Config classes for nessie client."""
from io import StringIO

import confuse
import yaml

from pynessie.conf.config_command import process
from pynessie.conf.config_parser import build_config
from pynessie.conf.io import write_to_file

__all__ = ["build_config", "to_dict", "write_to_file", "process"]


def to_dict(config: confuse.Configuration) -> dict:
    """Convert confuse Configuration object to dict."""
    yaml_config = config.dump(redact=True)
    dct = yaml.safe_load(StringIO(yaml_config))
    return dct



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Contents Commit Command CLI."""
from json import JSONDecodeError
from typing import List, Optional

import click

from pynessie.cli_common_context import ContextObject
from pynessie.client import NessieClient
from pynessie.decorators import error_handler, pass_client, validate_reference
from pynessie.error import NessieCliError, NessieNotFoundException
from pynessie.model import Content, ContentKey, ContentSchema, Delete, Operation, Put
from pynessie.types import CONTENT_KEY


@click.command("commit")
@click.option("-r", "--ref", help="Branch to commit content to. If not supplied the default branch from config is used.")
@click.option("-i", "--stdin", is_flag=True, help="Read content for --set from STDIN (separated by Ctrl-D)")
@click.option(
    "-s",
    "--expect-same-content",
    is_flag=True,
    help="Send the same content both as the new and expected (old contents) parameters for the edit operations",
)
@click.option(
    "-c",
    "--condition",
    "expected_hash",
    required=True,
    help="Expected hash. Only perform the action if the branch currently points to the hash specified by this option.",
)
@click.option("-R", "--remove-key", "delete", is_flag=True, help="Delete a content.")
@click.option("-m", "--message", help="Commit message.")
@click.option("--author", help="The author to use for the commit")
@click.argument("key", nargs=-1, required=True, type=CONTENT_KEY)
@pass_client
@error_handler
@validate_reference
def commit(
    ctx: ContextObject,
    ref: str,
    stdin: bool,
    expect_same_content: bool,
    expected_hash: str,
    delete: bool,
    message: str,
    author: str,
    key: List[ContentKey],
) -> None:
    """Commit content.

        KEY is the content key that is associated with a specific content to commit.
    This accepts as well multiple keys with space in between. The key can be in this format:
    'table.key' or 'namespace."table.key"'.

    Examples:

        nessie content commit -m "my awesome commit message" -r dev -c 122345abcd my_table -> Commit to branch 'dev'
    on expected hash '122345abcd' (head of the branch) the content from the interactive editor with the
    commit message "my awesome commit message" for content "my_table".

        nessie content commit -i -r dev -c 122345abcd my_table -> Commit to branch 'dev'
    on expected hash '122345abcd' (head of the branch) the content from STDIN with the commit message from
    the interactive editor for content "my_table".

        nessie content commit -m "my awesome commit message" -R -r dev -c 122345abcd my_table_1 my_table_2 ->
    Delete contents "my_table_1" and "my_table_2" from branch 'dev' on expected hash '122345abcd' and commit
    to branch 'dev" with the commit message "my awesome commit message".
    """
    operations = _get_commit_operations(ctx.nessie, ref, key, delete, stdin, expect_same_content)

    if operations:
        ctx.nessie.commit(ref, expected_hash, _get_message(message, key), author, *operations)
        click.echo("Changes have been committed.")
    else:
        click.echo("No changes requested")


def _get_commit_operations(
    client: NessieClient, ref: str, keys: List[ContentKey], is_delete: bool, use_stdin: bool = False, expect_same_contents: bool = False
) -> List[Operation]:
    if is_delete:
        return _get_delete_content_operations(keys)

    return _get_edit_content_operations(client, ref, keys, use_stdin, expect_same_contents)


def _get_delete_content_operations(keys: List[ContentKey]) -> List[Operation]:
    contents_operations: List[Operation] = []
    for content_key in keys:
        contents_operations.append(Delete(content_key))

    return contents_operations


def _get_edit_content_operations(
    client: NessieClient, ref: str, keys: List[ContentKey], use_stdin: bool = False, expect_same_contents: bool = False
) -> List[Operation]:
    contents_operations: List[Operation] = []
    for content_key in keys:
        content_original = _get_existing_content(client, ref, content_key)
        formatted_content_key = content_key.to_string()

        content_json = _get_edited_json_content(content_key, content_original, use_stdin)

        if _shall_skip_key(content_json, content_original):
            click.echo("Skipping key " + formatted_content_key + " (content not edited)")
            continue

        if _is_delete_operation(content_json, content_original):
            contents_operations.append(Delete(content_key))
            click.echo("Deleting content for key " + formatted_content_key)
            continue

        # If the user has an invalid JSON, we raise an Error
        try:
            content_new = ContentSchema().loads(content_json)
        except JSONDecodeError as e:
            raise NessieCliError("Invalid JSON Error", e.msg) from e

        # if we have the same content as the original one, we just skip
        if content_new == content_original:
            click.echo("Skipping key " + formatted_content_key + " (content not edited)")
            continue

        click.echo("Setting content for key " + formatted_content_key)
        if content_new.requires_expected_state():
            content_expected = content_new if expect_same_contents else content_original
            contents_operations.append(Put(content_key, content_new, content_expected))
        else:
            contents_operations.append(Put(content_key, content_new))

    return contents_operations


def _get_existing_content(client: NessieClient, ref: str, key: ContentKey) -> Optional[Content]:
    # get current contents, if present
    try:
        return client.get_content(ref, key)
    except NessieNotFoundException:
        return None


def _get_edited_json_content(key: ContentKey, content_orig: Optional[Content], use_stdin: bool) -> Optional[str]:
    if use_stdin:
        click.echo("Enter JSON content for key " + key.to_string())
        return click.get_text_stream("stdin").read().strip("\n")

    # allow the user to provide interactive input via the shell $EDITOR
    edited_content = _edit_contents(key.to_string(), (ContentSchema().dumps(content_orig) if content_orig else ""))

    if edited_content is not None:
        return edited_content.strip()

    return edited_content


def _edit_contents(key: str, body: str) -> Optional[str]:
    marker = "# Everything below is ignored\n"
    edit_message = (
        body
        + "\n\n"
        + marker
        + "Edit the content above to commit changes."
        + " Closing without change will result in a no-op."
        + "\nContent Key: "
        + key
    )
    try:
        message = click.edit(edit_message)
        # Note: None can occur when the user quits the editor without saving
        if message is None:
            return None
        message_altered = message.split(marker, 1)[0].strip("\n")
        return message_altered
    except click.ClickException:
        pass
    return None


def _shall_skip_key(content_json: Optional[str], content_original: Optional[Content]) -> bool:
    if content_json is None:
        # Skip the key if user for example exit with no editing e.g: :q!
        return True

    if not content_json and content_original is None:
        # Skip if no content (empty text) were edited and no existing content
        return True

    return False


def _is_delete_operation(content_json: Optional[str], content_original: Optional[Content]) -> bool:
    if not content_json and content_original is not None:
        # We let the user to remove content from existing content key which will be treated as DELETE op
        return True

    return False


def _get_message(message: str, keys: List[ContentKey]) -> Optional[str]:
    if message:
        return message
    return _get_commit_message(keys)


def _get_commit_message(keys: List[ContentKey]) -> Optional[str]:
    marker = "# Everything below is ignored\n"
    message = click.edit("\n\n" + marker + "Edit the commit message above: " + "\n".join([i.to_string() for i in keys]))
    return message.split(marker, 1)[0].rstrip("\n") if message is not None else None



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""log CLI command."""

import datetime
from typing import Any, List, Optional, Tuple

import click
from dateutil.tz import tzlocal

from pynessie.cli_common_context import ContextObject, MutuallyExclusiveOption
from pynessie.decorators import error_handler, pass_client, validate_reference
from pynessie.model import (
    CommitMetaSchema,
    LogEntry,
    LogEntrySchema,
    split_into_reference_and_hash,
)
from pynessie.utils import build_filter_for_commit_log_flags


@click.command("log")
@click.argument("ref", nargs=1, required=False)
@click.option("-n", "--number", help="number of log entries to return", type=int)
@click.option("--since", "--after", help="Only include commits newer than specific date, such as '2001-01-01T00:00:00+00:00'")
@click.option("--until", "--before", help="Only include commits older than specific date, such as '2999-12-30T23:00:00+00:00'")
@click.option(
    "--author",
    multiple=True,
    help="Limit commits to a specific author (this is the original committer). Supports specifying multiple authors to filter by.",
)
@click.option(
    "--committer",
    multiple=True,
    help="Limit commits to a specific committer (this is the logged in user/account who performed the commit). "
    "Supports specifying multiple committers to filter by.",
)
@click.option(
    "-r",
    "--revision-range",
    help="Hash to start viewing log from. If of the form '<start_hash>'..'<end_hash>' "
    "only show log for given range on the particular ref that was provided, the '<end_hash>' "
    "is inclusive and '<start_hash>' is exclusive.",
)
@click.option(
    "--filter",
    "query_filter",
    multiple=False,
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["author", "committer", "since", "until"],
    help="Allows advanced filtering using the Common Expression Language (CEL). "
    "An intro to CEL can be found at https://github.com/google/cel-spec/blob/master/doc/intro.md.\n"
    "Some examples with usable variables 'commit.author' (string) / 'commit.committer' (string) / 'commit.commitTime' (timestamp) / "
    "'commit.hash' (string) / 'commit.message' (string) / 'commit.properties' (map) are:\n"
    "commit.author=='nessie_author'\n"
    "commit.committer=='nessie_committer'\n"
    "timestamp(commit.commitTime) > timestamp('2021-06-21T10:39:17.977922Z')\n",
)
@click.option(
    "-x",
    "--extended",
    "fetch_all",
    is_flag=True,
    help="Retrieve all available information for the commit entries. "
    "This option will also return the operations for each commit and the parent hash. "
    "The schema of the JSON output will then produce a list of LogEntrySchema, otherwise a list of CommitMetaSchema.",
)
@pass_client
@error_handler
@validate_reference
def log(  # noqa: C901
    ctx: ContextObject,
    ref: str,
    number: int,
    since: str,
    until: str,
    author: List[str],
    committer: List[str],
    revision_range: str,
    query_filter: str,
    fetch_all: bool,
) -> None:
    """Show commit log.

    REF name of branch or tag to use to show the commit logs

    Examples:

        nessie log -> show commit logs using the configured default branch

        nessie log 1234567890abcdef -> show commit logs starting at commit 1234567890abcdef

        nessie log dev -> show commit logs for 'dev' branch, starting at the most recent (HEAD) commit of 'dev'

        nessie log dev@1234567890abcdef -> show commit logs for 'dev' branch, starting at commit 1234567890abcdef in 'dev'

        nessie log -n 5 dev -> show commit logs for 'dev' branch limited by 5 commits

        nessie log --revision-range 12345678abcdef..12345678efghj dev -> show commit logs in range of hash
    '12345678abcdef' and '12345678efghj' in 'dev' branch

        nessie log --author nessie.user dev -> show commit logs for user 'nessie.user' in 'dev' branch

        nessie log --filter "commit.author == 'nessie_user2' || commit.author == 'non_existing'" dev ->
    show commit logs using query in 'dev' branch

        nessie log --after "2019-01-01T00:00:00+00:00" --before "2021-01-01T00:00:00+00:00" dev ->
    show commit logs between "2019-01-01T00:00:00+00:00" and "2021-01-01T00:00:00+00:00" in 'dev' branch

    """
    ref, start_hash, end_hash = _log_ref_and_hashes(ref, revision_range)

    filtering_args: Any = {}
    if start_hash:
        filtering_args["startHash"] = start_hash
    if end_hash:
        filtering_args["endHash"] = end_hash
    expr = build_filter_for_commit_log_flags(query_filter, author, committer, since, until)
    if expr:
        filtering_args["filter"] = expr

    # TODO: limiting by path is not yet supported.
    log_result = ctx.nessie.get_log(start_ref=ref, max_records=number, fetch_all=fetch_all, **filtering_args)
    if ctx.json:
        if fetch_all:
            click.echo(LogEntrySchema().dumps(log_result, many=True))
        else:
            commit_metas = [entry.commit_meta for entry in log_result]
            click.echo(CommitMetaSchema().dumps(commit_metas, many=True))
    else:
        click.echo_via_pager(_format_log_result(x, ref, index, fetch_all) for index, x in enumerate(log_result))


def _log_ref_and_hashes(ref: str, revision_range: str) -> Tuple[str, Optional[str], Optional[str]]:
    start_hash = None
    ref, end_hash = split_into_reference_and_hash(ref)
    if revision_range:
        if ".." in revision_range:
            start_hash, end_hash_range = revision_range.split("..")
            if len(start_hash) == 0:
                start_hash = None
            if len(end_hash_range) > 0:
                if end_hash and end_hash != end_hash_range:
                    raise click.exceptions.BadOptionUsage(
                        "revision_range", "end-hash provided via 'ref' argument and 'revision-range', use only one of those"
                    )
                end_hash = end_hash_range
        elif end_hash and end_hash != revision_range:
            raise click.exceptions.BadOptionUsage(
                "revision_range", "end-hash provided via 'ref' argument and 'revision-range', use only one of those"
            )
        else:
            end_hash = revision_range
    return ref, start_hash, end_hash


def _format_log_result(x: LogEntry, ref: str, index: int, fetch_all: bool) -> str:
    result = _format_commit_log_string(x.commit_meta.hash_, ref, index)
    result += click.style(f"Author: {x.commit_meta.author}\n")
    result += click.style(f"Date: {_format_time(x.commit_meta.commitTime)}\n")
    if fetch_all:
        result += click.style(f"Parent: {x.parent_commit_hash}\n")
        result += click.style("Operations:\n")
        if not x.operations or len(x.operations) == 0:
            result += click.style("  (None)\n")
        else:
            for operation in x.operations:
                result += click.style(f"  {operation.pretty_print()}\n")
    result += click.style(f"\n\t{x.commit_meta.message}\n\n")
    return result


def _format_commit_log_string(commit_hash: str, ref: str, index: int) -> str:
    result = click.style(f"commit {commit_hash}", fg="yellow")
    if index == 0:
        result += click.style(f" ({ref})\n", fg="green")
    else:
        result += "\n"
    return result


def _format_time(date_time: datetime.datetime) -> str:
    return date_time.astimezone(tzlocal()).strftime("%c %z")



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""config CLI command."""

import click

from pynessie.cli_common_context import (
    ContextObject,
    DefaultHelp,
    MutuallyExclusiveOption,
)
from pynessie.conf import process
from pynessie.decorators import error_handler, pass_client


@click.command("config", cls=DefaultHelp)
@click.option("--get", cls=MutuallyExclusiveOption, help="get config parameter", mutually_exclusive=["set", "is_list", "unset"])
@click.option("--add", cls=MutuallyExclusiveOption, help="set config parameter", mutually_exclusive=["get", "is_list", "unset"])
@click.option(
    "-l",
    "--list",
    "is_list",
    cls=MutuallyExclusiveOption,
    is_flag=True,
    help="list config parameters",
    mutually_exclusive=["set", "get", "unset"],
)
@click.option("--unset", cls=MutuallyExclusiveOption, help="unset config parameter", mutually_exclusive=["get", "is_list", "set"])
@click.option("--type", "input_type", help="type to interpret config value to set or get. Allowed options: bool, int")
@click.argument("key", nargs=1, required=False)
@pass_client
@error_handler
def config(ctx: ContextObject, get: str, add: str, is_list: bool, unset: str, input_type: str, key: str) -> None:
    """Set and view config."""
    res = process(get, add, is_list, unset, key, input_type)
    click.echo(res)



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Top-level package for Python API and CLI for Nessie."""
import os
from typing import Optional

import confuse

from pynessie.client import NessieClient
from pynessie.conf import build_config

__author__ = """Project Nessie"""
__email__ = "nessie-release-builder@dremio.com"
__version__ = "0.62.1"


def get_config(config_dir: Optional[str] = None, args: Optional[dict] = None) -> confuse.Configuration:
    """Retrieve a confuse Configuration object."""
    if config_dir:
        os.environ["NESSIE_CLIENTDIR"] = config_dir
    return build_config(args)


def init(config_dir: Optional[str] = None, config_dict: Optional[dict] = None) -> NessieClient:
    """Create a new Nessie client object.

    :param config_dir: optional directory to look for config in
    :param config_dict: dictionary of extra config arguments
    :return: either a simple or rich client
    :example:
    >>> client = init('/my/config/dir')
    """
    if config_dict is None:
        config_dict = {}
    config = get_config(config_dir, args=config_dict)
    return _connect(config)


def _connect(config: confuse.Configuration) -> NessieClient:
    return NessieClient(config)



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Execute config command from cli."""
from typing import Optional, Union, cast

import click
import confuse

from pynessie.conf.config_parser import build_config
from pynessie.conf.io import write_to_file


def process(
    get_op: Optional[str],
    add_op: Optional[str],
    list_op: bool,
    unset_op: Optional[str],
    key: Optional[str],
    type_str: Optional[str] = "str",
) -> str:
    """Perform nessie config operations."""
    if (get_op or list_op or unset_op) and key:
        raise click.UsageError("can't set argument and specify option flags")
    if get_op or (key and not add_op):
        if not key and not get_op:
            raise click.UsageError("to show a key either the --get option or an argument must be supplied")
        config = _get_key(cast(str, get_op or key))
        return str(config.get(_get_type(type_str)))
    if add_op:
        config = build_config()
        config.set_args({add_op: _set_type(cast(str, key), type_str)}, dots=True)
        write_to_file(config)
        return ""
    if unset_op:
        config = build_config()
        for k in unset_op.split("."):
            config = config[k]
        config.set(None)
        write_to_file(config.root())
        return ""
    if list_op:
        config = build_config()
        return config.dump(redact=True)
    raise click.UsageError("no options set for config")


def _get_type(type_str: Optional[str]) -> type:
    if type_str == "bool":
        return bool
    if type_str == "int":
        return int

    return str


def _set_type(key: str, type_str: Optional[str]) -> Union[bool, int, str]:
    if type_str == "bool":
        return key.lower() in {"true", "1", "t", "y", "yes", "yeah", "yup", "certainly", "uh-huh"}
    if type_str == "int":
        return int(key)

    return key


def _get_key(key: str) -> confuse.Configuration:
    config = build_config()
    for i in key.split("."):
        if not i:
            continue
        config = config[i]
    return config



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Direct API operations on Nessie with requests."""

import os
from typing import Any, Optional, Union, cast

import requests
import simplejson as jsonlib
from requests.auth import AuthBase

from pynessie.error import _create_exception
from pynessie.model import ContentKey

DEFAULT_TIMEOUT_SEC = int(os.getenv("PYNESSIE_HTTP_TIMEOUT_SEC", "60"))


def _sanitize_timeout(timeout_sec: Optional[int]) -> Optional[int]:
    if timeout_sec is None:
        timeout_sec = DEFAULT_TIMEOUT_SEC
    if timeout_sec > 0:
        return timeout_sec
    return None  # no timeout


def _get_headers(has_body: bool = False) -> dict:
    headers = {"Accept": "application/json"}
    if has_body:
        headers = {"Content-Type": "application/json"}
    return headers


def _get(
    url: str, auth: Optional[AuthBase], ssl_verify: bool = True, params: Optional[dict] = None, timeout_sec: Optional[int] = None
) -> Union[str, dict, list]:
    timeout_sec = _sanitize_timeout(timeout_sec)
    r = requests.get(url, headers=_get_headers(), verify=ssl_verify, params=params, auth=auth, timeout=timeout_sec)
    return _check_error(r)


def _post(
    url: str,
    auth: Optional[AuthBase],
    json: Union[str, dict, None] = None,
    ssl_verify: bool = True,
    params: Optional[dict] = None,
    timeout_sec: Optional[int] = None,
) -> Union[str, dict, list]:
    timeout_sec = _sanitize_timeout(timeout_sec)
    if isinstance(json, str):
        json = jsonlib.loads(json)
    r = requests.post(
        url, headers=_get_headers(json is not None), verify=ssl_verify, json=json, params=params, auth=auth, timeout=timeout_sec
    )
    return _check_error(r)


def _delete(
    url: str, auth: Optional[AuthBase], ssl_verify: bool = True, params: Optional[dict] = None, timeout_sec: Optional[int] = None
) -> Union[str, dict, list]:
    timeout_sec = _sanitize_timeout(timeout_sec)
    r = requests.delete(url, headers=_get_headers(), verify=ssl_verify, params=params, auth=auth, timeout=timeout_sec)
    return _check_error(r)


def _put(
    url: str,
    auth: Optional[AuthBase],
    json: Union[str, dict, None] = None,
    ssl_verify: bool = True,
    params: Optional[dict] = None,
    timeout_sec: Optional[int] = None,
) -> Any:
    timeout_sec = _sanitize_timeout(timeout_sec)
    if isinstance(json, str):
        json = jsonlib.loads(json)
    r = requests.put(
        url, headers=_get_headers(json is not None), verify=ssl_verify, json=json, params=params, auth=auth, timeout=timeout_sec
    )
    return _check_error(r)


def _check_error(r: requests.models.Response) -> Union[dict, list]:
    if 200 <= r.status_code < 300:
        return r.json() if r.content else {}

    if isinstance(r.reason, bytes):
        try:
            reason = r.reason.decode("utf-8")
        except UnicodeDecodeError:
            reason = r.reason.decode("iso-8859-1")
    else:
        reason = r.reason

    try:
        parsed_response = r.json()
    except:  # NOQA # pylint: disable=W0702
        # rare/unexpected case when the server responds with a non-JSON payload for an error
        parsed_response = {}

    raise _create_exception(parsed_response, r.status_code, reason, r.url)


def all_references(base_url: str, auth: Optional[AuthBase], ssl_verify: bool = True, fetch_all: bool = False) -> dict:
    """Fetch all known references.

    :param base_url: base Nessie url
    :param auth: Authentication settings
    :param ssl_verify: ignore ssl errors if False
    :param fetch_all: indicates whether additional metadata should be fetched
    :return: json list of Nessie references
    """
    params = {}
    if fetch_all:
        params["fetch"] = "ALL"
    return cast(dict, _get(base_url + "/trees", auth, ssl_verify=ssl_verify, params=params))


def get_reference(base_url: str, auth: Optional[AuthBase], ref: str, ssl_verify: bool = True) -> dict:
    """Fetch a reference.

    :param base_url: base Nessie url
    :param auth: Authentication settings
    :param ref: name of ref to fetch
    :param ssl_verify: ignore ssl errors if False
    :return: json Nessie branch or tag
    """
    return cast(dict, _get(base_url + "/trees/tree/{}".format(ref), auth, ssl_verify=ssl_verify))


def create_reference(
    base_url: str, auth: Optional[AuthBase], ref_json: dict, source_ref: Optional[str] = None, ssl_verify: bool = True
) -> dict:
    """Create a reference.

    :param base_url: base Nessie url
    :param auth: Authentication settings
    :param ref_json: reference to create as json object
    :param source_ref: name of the reference via which the hash in 'ref_json' is reachable
    :param ssl_verify: ignore ssl errors if False
    :return: json Nessie branch or tag
    """
    params = {}
    if source_ref:
        params["sourceRefName"] = source_ref
    return cast(dict, _post(base_url + "/trees/tree", auth, ref_json, ssl_verify=ssl_verify, params=params))


def get_default_branch(base_url: str, auth: Optional[AuthBase], ssl_verify: bool = True) -> dict:
    """Fetch a reference.

    :param base_url: base Nessie url
    :param auth: Authentication settings
    :param ssl_verify: ignore ssl errors if False
    :return: json Nessie branch
    """
    return cast(dict, _get(base_url + "/trees/tree", auth, ssl_verify=ssl_verify))


def delete_branch(base_url: str, auth: Optional[AuthBase], branch: str, hash_: str, ssl_verify: bool = True) -> None:
    """Delete a branch.

    :param base_url: base Nessie url
    :param auth: Authentication settings
    :param branch: name of branch to delete
    :param hash_: branch hash
    :param ssl_verify: ignore ssl errors if False
    """
    params = {"expectedHash": hash_}
    _delete(base_url + "/trees/branch/{}".format(branch), auth, ssl_verify=ssl_verify, params=params)


def delete_tag(base_url: str, auth: Optional[AuthBase], tag: str, hash_: str, ssl_verify: bool = True) -> None:
    """Delete a tag.

    :param base_url: base Nessie url
    :param auth: Authentication settings
    :param tag: name of tag to delete
    :param hash_: tag hash
    :param ssl_verify: ignore ssl errors if False
    """
    params = {"expectedHash": hash_}
    _delete(base_url + "/trees/tag/{}".format(tag), auth, ssl_verify=ssl_verify, params=params)


def list_tables(
    base_url: str,
    auth: Optional[AuthBase],
    ref: str,
    hash_on_ref: Optional[str] = None,
    max_result_hint: Optional[int] = None,
    page_token: Optional[str] = None,
    query_filter: Optional[str] = None,
    ssl_verify: bool = True,
) -> list:
    """Fetch a list of all tables from a known reference.

    :param base_url: base Nessie url
    :param auth: Authentication settings
    :param ref: reference
    :param hash_on_ref: hash on reference
    :param max_result_hint: hint for the server, maximum number of results to return
    :param page_token: the token retrieved from a previous page returned for the same ref
    :param query_filter: A CEL expression that allows advanced filtering capabilities
    :param ssl_verify: ignore ssl errors if False
    :return: json list of Nessie table names
    """
    params = {}
    if max_result_hint:
        params["maxRecords"] = str(max_result_hint)
    if hash_on_ref:
        params["hashOnRef"] = hash_on_ref
    if page_token:
        params["pageToken"] = page_token
    if query_filter:
        params["filter"] = query_filter
    return cast(list, _get(base_url + "/trees/tree/{}/entries".format(ref), auth, ssl_verify=ssl_verify, params=params))


def list_logs(
    base_url: str,
    auth: Optional[AuthBase],
    ref: str,
    hash_on_ref: Optional[str] = None,
    ssl_verify: bool = True,
    max_records: Optional[int] = None,
    fetch_all: bool = False,
    **filtering_args: Any,
) -> dict:
    """Fetch a list of all logs from a known starting reference.

    :param base_url: base Nessie url
    :param auth: Authentication settings
    :param ref: starting reference
    :param hash_on_ref: hash on reference
    :param max_records: maximum number of entries to return
    :param ssl_verify: ignore ssl errors if False
    :param fetch_all: indicates whether additional metadata should be fetched
    :param filtering_args: All of the args used to filter the log
    :return: json dict of Nessie logs
    """
    params = filtering_args
    if hash_on_ref:
        params["hashOnRef"] = hash_on_ref
    if max_records:
        params["maxRecords"] = max_records
    if fetch_all:
        params["fetch"] = "ALL"
    return cast(dict, _get(base_url + "/trees/tree/{}/log".format(ref), auth, ssl_verify=ssl_verify, params=filtering_args))


def get_content(
    base_url: str, auth: Optional[AuthBase], ref: str, content_key: ContentKey, hash_on_ref: Optional[str] = None, ssl_verify: bool = True
) -> dict:
    """Fetch a table from a known branch.

    :param base_url: base Nessie url
    :param auth: Authentication settings
    :param ref: ref
    :param hash_on_ref: hash on reference
    :param content_key: key that is associated with content like table
    :param ssl_verify: ignore ssl errors if False
    :return: json dict of Nessie table
    """
    params = {"ref": ref}
    if hash_on_ref:
        params["hashOnRef"] = hash_on_ref
    return cast(dict, _get(base_url + "/contents/{}".format(content_key.to_path_string()), auth, ssl_verify=ssl_verify, params=params))


def assign_branch(
    base_url: str, auth: Optional[AuthBase], branch: str, assign_to_json: dict, old_hash: Optional[str], ssl_verify: bool = True
) -> None:
    """Assign a reference to a branch.

    :param base_url: base Nessie url
    :param auth: Authentication settings
    :param branch: name of the branch
    :param assign_to_json: hash to become the new HEAD of the branch and the name of the reference via which that hash is reachable
    :param old_hash: current hash of the branch
    :param ssl_verify: ignore ssl errors if False
    """
    url = "/trees/branch/{}".format(branch)
    params = {"expectedHash": old_hash}
    _put(base_url + url, auth, assign_to_json, ssl_verify=ssl_verify, params=params)


def assign_tag(
    base_url: str, auth: Optional[AuthBase], tag: str, assign_to_json: dict, old_hash: Optional[str], ssl_verify: bool = True
) -> None:
    """Assign a reference to a tag.

    :param base_url: base Nessie url
    :param auth: Authentication settings
    :param tag: name of the tag
    :param assign_to_json: hash to become the new HEAD of the tag and the name of the reference via which that hash is reachable
    :param old_hash: current hash of the tag
    :param ssl_verify: ignore ssl errors if False
    """
    url = "/trees/tag/{}".format(tag)
    params = {"expectedHash": old_hash}
    _put(base_url + url, auth, assign_to_json, ssl_verify=ssl_verify, params=params)


def cherry_pick(
    base_url: str, auth: Optional[AuthBase], branch: str, transplant_json: dict, expected_hash: Optional[str], ssl_verify: bool = True
) -> None:
    """cherry-pick a list of hashes to a branch.

    :param base_url: base Nessie url
    :param auth: Authentication settings
    :param branch: name of branch to cherry pick onto
    :param transplant_json: transplant content
    :param expected_hash: expected hash of HEAD of branch
    :param ssl_verify: ignore ssl errors if False
    """
    url = "/trees/branch/{}/transplant".format(branch)
    params = {}
    if expected_hash:
        params["expectedHash"] = expected_hash
    _post(base_url + url, auth, json=transplant_json, ssl_verify=ssl_verify, params=params)


def merge(
    base_url: str, auth: Optional[AuthBase], branch: str, merge_json: dict, expected_hash: Optional[str], ssl_verify: bool = True
) -> dict:
    """Merge a branch into another branch.

    :param base_url: base Nessie url
    :param auth: Authentication settings
    :param branch: name of branch to merge onto
    :param merge_json: merge content
    :param expected_hash: expected hash of HEAD of branch
    :param ssl_verify: ignore ssl errors if False
    :return: json dict of a merge response
    """
    url = "/trees/branch/{}/merge".format(branch)
    params = {}
    if expected_hash:
        params["expectedHash"] = expected_hash
    response = _post(base_url + url, auth, json=merge_json, ssl_verify=ssl_verify, params=params)
    return cast(dict, response)


def commit(
    base_url: str,
    auth: Optional[AuthBase],
    branch: str,
    operations: str,
    expected_hash: str,
    ssl_verify: bool = True,
) -> dict:
    """Commit a set of operations to a branch.

    :param base_url: base Nessie url
    :param auth: Authentication settings
    :param branch: name of branch to merge onto
    :param operations: json object of operations
    :param expected_hash: expected hash of HEAD of branch
    :param ssl_verify: ignore ssl errors if False
    """
    url = "/trees/branch/{}/commit".format(branch)
    params = {"expectedHash": expected_hash}
    return cast(dict, _post(base_url + url, auth, json=operations, ssl_verify=ssl_verify, params=params))


def get_diff(
    base_url: str,
    auth: Optional[AuthBase],
    from_ref: str,
    to_ref: str,
    from_hash_on_ref: Optional[str] = None,
    to_hash_on_ref: Optional[str] = None,
    ssl_verify: bool = True,
) -> dict:
    """Fetch the diff for two given references.

    :param base_url: base Nessie url
    :param auth: Authentication settings
    :param from_ref: the starting ref for the diff
    :param to_ref:  the ending ref for the diff
    :param from_hash_on_ref: optional hash on from reference
    :param to_hash_on_ref: optional hash on to reference
    :param ssl_verify: ignore ssl errors if False
    :return: json dict of a Diff
    """
    from_hash_on_ref_asterisk = f"*{from_hash_on_ref}" if from_hash_on_ref else ""
    to_hash_on_ref_asterisk = f"*{to_hash_on_ref}" if to_hash_on_ref else ""
    return cast(
        dict,
        _get(f"{base_url}/diffs/{from_ref}{from_hash_on_ref_asterisk}...{to_ref}{to_hash_on_ref_asterisk}", auth, ssl_verify=ssl_verify),
    )


def list_reflog(
    base_url: str,
    auth: Optional[AuthBase],
    ssl_verify: bool = True,
    max_records: Optional[int] = None,
    **query_params: Any,
) -> dict:
    """Fetch a list of all reflog from a head or from a specified range.

    :param base_url: base Nessie url
    :param auth: Authentication settings
    :param max_records: maximum number of entries to return
    :param ssl_verify: ignore ssl errors if False
    :param query_params: All of the args used to filter the log
    :return: json dict of Nessie logs
    """
    params = query_params
    if max_records:
        params["maxRecords"] = max_records
    return cast(dict, _get(f"{base_url}/reflogs", auth, ssl_verify=ssl_verify, params=query_params))



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Tests for `pynessie` package."""
import pytest

from pynessie import init
from pynessie.error import NessieConflictException
from pynessie.model import Branch, Entries


@pytest.mark.vcr
def test_client_interface_e2e() -> None:
    """Test client object against live server."""
    client = init()
    assert isinstance(client.get_base_url(), str)
    assert client.get_base_url() == "http://localhost:19120/api/v1"
    references = client.list_references().references
    assert len(references) == 1
    assert references[0] == Branch("main", references[0].hash_)
    main_name = references[0].name
    main_commit = references[0].hash_
    with pytest.raises(NessieConflictException):
        client.create_branch("main")
    created_reference = client.create_branch("test", main_name, main_commit)
    references = client.list_references().references
    assert len(references) == 2
    assert next(i for i in references if i.name == "main") == Branch("main", main_commit)
    assert next(i for i in references if i.name == "test") == Branch("test", main_commit)
    reference = client.get_reference("test")
    assert created_reference == reference
    assert isinstance(reference.hash_, str)
    tables = client.list_keys(reference.name, reference.hash_)
    assert isinstance(tables, Entries)
    assert len(tables.entries) == 0
    assert isinstance(main_commit, str)
    client.delete_branch("test", main_commit)
    references = client.list_references().references
    assert len(references) == 1



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Nessie Exceptions."""
from typing import Optional

import simplejson as json


class NessieInvalidUsageException(Exception):
    """Exception for invalid usage."""

    def __init__(self, msg: Optional[str] = None) -> None:
        """Construct NessieInvalidUsageException."""
        super().__init__(msg)


class NessieException(Exception):
    """Base Nessie exception."""

    def __init__(
        self,
        parsed_response: dict,
        status: int,
        url: str,
        reason: str,
        msg: Optional[str] = None,
    ) -> None:
        """Construct base Nessie Exception."""
        if "message" in parsed_response:
            exception_msg = reason + ": " + parsed_response["message"]
        elif msg:
            exception_msg = reason + ": " + msg
        else:
            # If we did not get a JSON response for a 500 error, use a generic message.
            # In this case the HTML response contents are likely not user-friendly anyway.
            if 500 <= status <= 599:
                exception_msg = "Internal Server Error"
            else:
                exception_msg = "Unknown Error"

        super().__init__(exception_msg)

        self.status_code = status
        self.server_message = parsed_response.get("message", str(parsed_response))
        self.server_status = parsed_response.get("status", "UNKNOWN")
        self.error_code = parsed_response.get("errorCode", "UNKNOWN")
        self.server_stack_trace = parsed_response.get("serverStackTrace")
        self.url = url

    def json(self) -> str:
        """Dump this exception as a json object."""
        return json.dumps(
            {
                "server_message": self.server_message,
                "error_code": self.error_code,
                "server_status": self.server_status,
                "server_stack_trace": self.server_stack_trace,
                "status_code": self.status_code,
                "url": self.url,
                "msg": " ".join(self.args),
            }
        )


class NessieUnauthorizedException(NessieException):
    """Nessie exception for login error 401."""

    pass


class NessiePermissionException(NessieException):
    """Nessie exception for permission error 403."""

    pass


class NessieNotFoundException(NessieException):
    """Generic Nessie exception for not found error 404."""

    pass


class NessieReferenceNotFoundException(NessieNotFoundException):
    """This exception is thrown when a requested reference is not present in the store."""

    pass


class NessieContentNotFoundException(NessieNotFoundException):
    """This exception is thrown when the requested content object is not present in the store."""

    pass


class NessiePreconidtionFailedException(NessieException):
    """Nessie exception for pre-condition failed error 412."""

    pass


class NessieConflictException(NessieException):
    """Generic Nessie exception for conflict error 409."""

    pass


class NessieReferenceAlreadyExistsException(NessieConflictException):
    """Reference not found.

    This exception is thrown when a reference could not be created because another reference with the same name is
    already present in the store.
    """

    pass


class NessieReferenceConflictException(NessieConflictException):
    """Expected hash did not match actual hash on a reference.

    This exception is thrown when the hash associated with a named reference does not match with the hash provided
    by the caller.
    """

    pass


class NessieServerException(NessieException):
    """Nessie exception for server errors 5xx."""

    pass


class NessieCliError(Exception):
    """Base Nessie CLI related errors."""

    def __init__(self, title: str, msg: Optional[str] = None) -> None:
        """Construct base Nessie CLI Error."""
        super().__init__()

        self.title = title
        self.msg = msg

    def json(self) -> str:
        """Dump this error as a json object."""
        return json.dumps(
            {
                "title": self.title,
                "message": self.msg,
            }
        )


def _create_nessie_exception(error: dict, status: int, reason: str, url: str) -> Optional[NessieException]:
    if "errorCode" not in error:
        return None

    error_code = error["errorCode"]
    if error_code == "REFERENCE_NOT_FOUND":
        return NessieReferenceNotFoundException(error, status, url, reason)
    if error_code == "REFERENCE_ALREADY_EXISTS":
        return NessieReferenceAlreadyExistsException(error, status, url, reason)
    if error_code == "CONTENT_NOT_FOUND":
        return NessieContentNotFoundException(error, status, url, reason)
    if error_code == "REFERENCE_CONFLICT":
        return NessieReferenceConflictException(error, status, url, reason)
    return None


def _create_exception(error: dict, status: int, reason: str, url: str) -> NessieException:
    if 400 <= status < 500:
        reason = f"Client Error {reason}"
    elif 500 <= status < 600:
        reason = f"Server Error {reason}"

    ex = _create_nessie_exception(error, status, reason, url)
    if ex:
        return ex

    if status == 412:
        return NessiePreconidtionFailedException(error, status, url, reason, msg="Unable to complete transaction, please retry.")
    if status == 401:
        return NessieUnauthorizedException(error, status, url, reason, msg="Unauthorized to access API endpoint")
    if status == 403:
        return NessiePermissionException(error, status, url, reason, msg="Insufficient permissions to access " + url)
    if status == 404:
        return NessieNotFoundException(error, status, url, reason, msg="Entity not found at " + url)
    if status == 409:
        return NessieConflictException(error, status, url, reason, msg="Unable to complete transaction.")
    if 500 <= status <= 599:
        return NessieServerException(error, status, url, reason)
    return NessieException(error, status, url, reason)



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""reflog CLI command."""

import datetime
from typing import Any

import click
from dateutil.tz import tzlocal

from pynessie.cli_common_context import ContextObject
from pynessie.decorators import error_handler, pass_client
from pynessie.model import ReflogEntry, ReflogEntrySchema


@click.command("reflog")
@click.option("-n", "--number", help="number of reflog entries to return", type=int)
@click.option(
    "-r",
    "--revision-range",
    help="Hash to start viewing reflog from. If of the form '<start_hash>'..'<end_hash>' "
    "only show reflog for given range on the particular ref that was provided. Both the '<end_hash>' "
    "and '<start_hash>' is inclusive.",
)
@pass_client
@error_handler
def reflog(  # noqa: C901
    ctx: ContextObject,
    number: int,
    revision_range: str,
) -> None:
    """Show reflog.

    Reflog entries from the current HEAD or from the specified revision_range.

    Examples:

        nessie reflog -> show reflog entries from the current HEAD

        nessie reflog -n 5 -> show reflog entries from the current HEAD limited by 5 entries

        nessie reflog --revision-range 12345678abcdef..12345678efghj -> show reflog entries in range of hash
    '12345678abcdef' and '12345678efghj' (both inclusive)

    """
    start_hash = None
    end_hash = None
    if revision_range:
        if ".." in revision_range:
            start_hash, end_hash = revision_range.split("..")
        else:
            end_hash = revision_range

    query_params: Any = {}
    if start_hash:
        query_params["startHash"] = start_hash
    if end_hash:
        query_params["endHash"] = end_hash

    log_result = ctx.nessie.get_reflog(max_records=number, **query_params)
    if ctx.json:
        click.echo(ReflogEntrySchema().dumps(log_result, many=True))
    else:
        click.echo_via_pager(_format_reflog_result(x) for x in log_result)


def _format_reflog_result(x: ReflogEntry) -> str:
    result = click.style(f"reflog {x.reflog_id}", fg="yellow")
    result += click.style(f" ({x.ref_type} : {x.ref_name})\n", fg="green")
    result += click.style(f"Operation: {x.operation}\n")
    result += click.style(f"OperationTime: {_format_time(x.operation_time)}\n")
    result += click.style(f"CommitHash: {x.commit_hash}\n")
    result += click.style(f"ParentReflogId: {x.parent_reflog_id}\n")
    result += click.style("SourceHashes:\n")
    if not x.source_hashes or len(x.source_hashes) == 0:
        result += click.style("  (None)\n")
    else:
        for source_hash in x.source_hashes:
            result += click.style(f"  {source_hash}\n")
    result += click.style("\n")
    return result


def _format_time(epoch_micro: int) -> str:
    epoch_seconds = epoch_micro / 1000000
    return datetime.datetime.fromtimestamp(epoch_seconds, tz=tzlocal()).strftime("%c %z")



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Bearer auth for Nessie client."""
from requests import models
from requests.auth import AuthBase


class TokenAuth(AuthBase):
    """AuthBase override for bearer token based auth."""

    def __init__(self, token: str) -> None:
        """Create an AuthBase for the specified bearer auth token."""
        self._token = token

    def __call__(self, r: models.PreparedRequest) -> models.PreparedRequest:
        """Append Auth Header to request."""
        r.headers["Authorization"] = "Bearer {}".format(self._token)
        return r



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Tests for `pynessie.auth` package."""

import os
from pathlib import Path

import requests
from assertpy import assert_that

from pynessie.auth.config import setup_auth
from pynessie.conf.config_parser import build_config


def test_auth_type_none() -> None:
    """Makes sure the auth type "none" resolves to a None authenticator."""
    config = build_config({"auth.type": "none"})
    assert_that(setup_auth(config)).is_none()


def test_bearer_token() -> None:
    """Makes sure the bearer auth token is added to HTTP request headers."""
    config = build_config({"auth.type": "bearer", "auth.token": "test_token_123"})
    auth = setup_auth(config)
    r = requests.Request(url="http://127.255.0.0:0/", auth=auth)
    prepared = r.prepare()
    assert_that(prepared.headers.get("Authorization")).is_equal_to("Bearer test_token_123")


def test_basic_auth() -> None:
    """Makes sure the basic auth token is added to HTTP request headers."""
    config = build_config({"auth.type": "basic", "auth.username": "user_123", "auth.password": "pwd_123"})
    auth = setup_auth(config)
    r = requests.Request(url="http://127.255.0.0:0/", auth=auth)
    prepared = r.prepare()
    assert_that(prepared.headers.get("Authorization")).is_equal_to("Basic dXNlcl8xMjM6cHdkXzEyMw==")


def test_aws_auth() -> None:
    """Makes sure the request is signed according to AWS auth requirements."""
    config = build_config({"auth.type": "aws", "auth.region": "us-west-2"})
    os.environ["AWS_ACCESS_KEY_ID"] = "test_key_id"
    os.environ["AWS_SECRET_ACCESS_KEY"] = "test_secret_key"  # noqa: S105
    auth = setup_auth(config)
    r = requests.Request(url="http://127.255.0.0:0/", method="GET", auth=auth)
    prepared = r.prepare()
    assert_that(prepared.headers.get("Authorization")).matches(
        r"AWS4-HMAC-SHA256 Credential=test_key_id/[0-9]+/us-west-2/execute-api/aws4_request, "
        r"SignedHeaders=host;x-amz-content-sha256;x-amz-date, Signature=.+"
    )


def test_aws_auth_profile(tmp_path: Path) -> None:
    """Makes sure the request is signed according to AWS auth requirements."""
    aws_credentials = tmp_path / "aws.config"
    aws_credentials.write_text(
        """
[test1]
aws_access_key_id=test1_key
aws_secret_access_key=test1_secret
aws_session_token=test1_session
    """
    )
    config = build_config({"auth.type": "aws", "auth.region": "us-west-2", "auth.profile": "test1"})
    os.environ["AWS_SHARED_CREDENTIALS_FILE"] = str(aws_credentials)
    auth = setup_auth(config)
    r = requests.Request(url="http://127.255.0.0:0/", method="GET", auth=auth)
    prepared = r.prepare()
    assert_that(prepared.headers.get("Authorization")).matches(
        r"AWS4-HMAC-SHA256 Credential=test1_key/[0-9]+/us-west-2/execute-api/aws4_request, "
        r"SignedHeaders=host;x-amz-content-sha256;x-amz-date;x-amz-security-token, Signature=.+"
    )
